# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: signer.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='signer.proto',
  package='signrpc',
  syntax='proto3',
  serialized_options=b'Z-github.com/lightningnetwork/lnd/lnrpc/signrpc',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0csigner.proto\x12\x07signrpc\"3\n\nKeyLocator\x12\x12\n\nkey_family\x18\x01 \x01(\x05\x12\x11\n\tkey_index\x18\x02 \x01(\x05\"L\n\rKeyDescriptor\x12\x15\n\rraw_key_bytes\x18\x01 \x01(\x0c\x12$\n\x07key_loc\x18\x02 \x01(\x0b\x32\x13.signrpc.KeyLocator\")\n\x05TxOut\x12\r\n\x05value\x18\x01 \x01(\x03\x12\x11\n\tpk_script\x18\x02 \x01(\x0c\"\xc4\x01\n\x0eSignDescriptor\x12(\n\x08key_desc\x18\x01 \x01(\x0b\x32\x16.signrpc.KeyDescriptor\x12\x14\n\x0csingle_tweak\x18\x02 \x01(\x0c\x12\x14\n\x0c\x64ouble_tweak\x18\x03 \x01(\x0c\x12\x16\n\x0ewitness_script\x18\x04 \x01(\x0c\x12\x1e\n\x06output\x18\x05 \x01(\x0b\x32\x0e.signrpc.TxOut\x12\x0f\n\x07sighash\x18\x07 \x01(\r\x12\x13\n\x0binput_index\x18\x08 \x01(\x05\"L\n\x07SignReq\x12\x14\n\x0craw_tx_bytes\x18\x01 \x01(\x0c\x12+\n\nsign_descs\x18\x02 \x03(\x0b\x32\x17.signrpc.SignDescriptor\"\x1c\n\x08SignResp\x12\x10\n\x08raw_sigs\x18\x01 \x03(\x0c\"2\n\x0bInputScript\x12\x0f\n\x07witness\x18\x01 \x03(\x0c\x12\x12\n\nsig_script\x18\x02 \x01(\x0c\">\n\x0fInputScriptResp\x12+\n\rinput_scripts\x18\x01 \x03(\x0b\x32\x14.signrpc.InputScript\"m\n\x0eSignMessageReq\x12\x0b\n\x03msg\x18\x01 \x01(\x0c\x12$\n\x07key_loc\x18\x02 \x01(\x0b\x32\x13.signrpc.KeyLocator\x12\x13\n\x0b\x64ouble_hash\x18\x03 \x01(\x08\x12\x13\n\x0b\x63ompact_sig\x18\x04 \x01(\x08\"$\n\x0fSignMessageResp\x12\x11\n\tsignature\x18\x01 \x01(\x0c\"B\n\x10VerifyMessageReq\x12\x0b\n\x03msg\x18\x01 \x01(\x0c\x12\x11\n\tsignature\x18\x02 \x01(\x0c\x12\x0e\n\x06pubkey\x18\x03 \x01(\x0c\"\"\n\x11VerifyMessageResp\x12\r\n\x05valid\x18\x01 \x01(\x08\"\x80\x01\n\x10SharedKeyRequest\x12\x18\n\x10\x65phemeral_pubkey\x18\x01 \x01(\x0c\x12(\n\x07key_loc\x18\x02 \x01(\x0b\x32\x13.signrpc.KeyLocatorB\x02\x18\x01\x12(\n\x08key_desc\x18\x03 \x01(\x0b\x32\x16.signrpc.KeyDescriptor\"\'\n\x11SharedKeyResponse\x12\x12\n\nshared_key\x18\x01 \x01(\x0c\x32\xd4\x02\n\x06Signer\x12\x34\n\rSignOutputRaw\x12\x10.signrpc.SignReq\x1a\x11.signrpc.SignResp\x12@\n\x12\x43omputeInputScript\x12\x10.signrpc.SignReq\x1a\x18.signrpc.InputScriptResp\x12@\n\x0bSignMessage\x12\x17.signrpc.SignMessageReq\x1a\x18.signrpc.SignMessageResp\x12\x46\n\rVerifyMessage\x12\x19.signrpc.VerifyMessageReq\x1a\x1a.signrpc.VerifyMessageResp\x12H\n\x0f\x44\x65riveSharedKey\x12\x19.signrpc.SharedKeyRequest\x1a\x1a.signrpc.SharedKeyResponseB/Z-github.com/lightningnetwork/lnd/lnrpc/signrpcb\x06proto3'
)




_KEYLOCATOR = _descriptor.Descriptor(
  name='KeyLocator',
  full_name='signrpc.KeyLocator',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key_family', full_name='signrpc.KeyLocator.key_family', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key_index', full_name='signrpc.KeyLocator.key_index', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25,
  serialized_end=76,
)


_KEYDESCRIPTOR = _descriptor.Descriptor(
  name='KeyDescriptor',
  full_name='signrpc.KeyDescriptor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='raw_key_bytes', full_name='signrpc.KeyDescriptor.raw_key_bytes', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key_loc', full_name='signrpc.KeyDescriptor.key_loc', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=78,
  serialized_end=154,
)


_TXOUT = _descriptor.Descriptor(
  name='TxOut',
  full_name='signrpc.TxOut',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='signrpc.TxOut.value', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pk_script', full_name='signrpc.TxOut.pk_script', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=156,
  serialized_end=197,
)


_SIGNDESCRIPTOR = _descriptor.Descriptor(
  name='SignDescriptor',
  full_name='signrpc.SignDescriptor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key_desc', full_name='signrpc.SignDescriptor.key_desc', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='single_tweak', full_name='signrpc.SignDescriptor.single_tweak', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='double_tweak', full_name='signrpc.SignDescriptor.double_tweak', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='witness_script', full_name='signrpc.SignDescriptor.witness_script', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='output', full_name='signrpc.SignDescriptor.output', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sighash', full_name='signrpc.SignDescriptor.sighash', index=5,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='input_index', full_name='signrpc.SignDescriptor.input_index', index=6,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=200,
  serialized_end=396,
)


_SIGNREQ = _descriptor.Descriptor(
  name='SignReq',
  full_name='signrpc.SignReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='raw_tx_bytes', full_name='signrpc.SignReq.raw_tx_bytes', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sign_descs', full_name='signrpc.SignReq.sign_descs', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=398,
  serialized_end=474,
)


_SIGNRESP = _descriptor.Descriptor(
  name='SignResp',
  full_name='signrpc.SignResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='raw_sigs', full_name='signrpc.SignResp.raw_sigs', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=476,
  serialized_end=504,
)


_INPUTSCRIPT = _descriptor.Descriptor(
  name='InputScript',
  full_name='signrpc.InputScript',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='witness', full_name='signrpc.InputScript.witness', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sig_script', full_name='signrpc.InputScript.sig_script', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=506,
  serialized_end=556,
)


_INPUTSCRIPTRESP = _descriptor.Descriptor(
  name='InputScriptResp',
  full_name='signrpc.InputScriptResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='input_scripts', full_name='signrpc.InputScriptResp.input_scripts', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=558,
  serialized_end=620,
)


_SIGNMESSAGEREQ = _descriptor.Descriptor(
  name='SignMessageReq',
  full_name='signrpc.SignMessageReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='msg', full_name='signrpc.SignMessageReq.msg', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key_loc', full_name='signrpc.SignMessageReq.key_loc', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='double_hash', full_name='signrpc.SignMessageReq.double_hash', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='compact_sig', full_name='signrpc.SignMessageReq.compact_sig', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=622,
  serialized_end=731,
)


_SIGNMESSAGERESP = _descriptor.Descriptor(
  name='SignMessageResp',
  full_name='signrpc.SignMessageResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='signature', full_name='signrpc.SignMessageResp.signature', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=733,
  serialized_end=769,
)


_VERIFYMESSAGEREQ = _descriptor.Descriptor(
  name='VerifyMessageReq',
  full_name='signrpc.VerifyMessageReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='msg', full_name='signrpc.VerifyMessageReq.msg', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signature', full_name='signrpc.VerifyMessageReq.signature', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pubkey', full_name='signrpc.VerifyMessageReq.pubkey', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=771,
  serialized_end=837,
)


_VERIFYMESSAGERESP = _descriptor.Descriptor(
  name='VerifyMessageResp',
  full_name='signrpc.VerifyMessageResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='valid', full_name='signrpc.VerifyMessageResp.valid', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=839,
  serialized_end=873,
)


_SHAREDKEYREQUEST = _descriptor.Descriptor(
  name='SharedKeyRequest',
  full_name='signrpc.SharedKeyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ephemeral_pubkey', full_name='signrpc.SharedKeyRequest.ephemeral_pubkey', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key_loc', full_name='signrpc.SharedKeyRequest.key_loc', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key_desc', full_name='signrpc.SharedKeyRequest.key_desc', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=876,
  serialized_end=1004,
)


_SHAREDKEYRESPONSE = _descriptor.Descriptor(
  name='SharedKeyResponse',
  full_name='signrpc.SharedKeyResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='shared_key', full_name='signrpc.SharedKeyResponse.shared_key', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1006,
  serialized_end=1045,
)

_KEYDESCRIPTOR.fields_by_name['key_loc'].message_type = _KEYLOCATOR
_SIGNDESCRIPTOR.fields_by_name['key_desc'].message_type = _KEYDESCRIPTOR
_SIGNDESCRIPTOR.fields_by_name['output'].message_type = _TXOUT
_SIGNREQ.fields_by_name['sign_descs'].message_type = _SIGNDESCRIPTOR
_INPUTSCRIPTRESP.fields_by_name['input_scripts'].message_type = _INPUTSCRIPT
_SIGNMESSAGEREQ.fields_by_name['key_loc'].message_type = _KEYLOCATOR
_SHAREDKEYREQUEST.fields_by_name['key_loc'].message_type = _KEYLOCATOR
_SHAREDKEYREQUEST.fields_by_name['key_desc'].message_type = _KEYDESCRIPTOR
DESCRIPTOR.message_types_by_name['KeyLocator'] = _KEYLOCATOR
DESCRIPTOR.message_types_by_name['KeyDescriptor'] = _KEYDESCRIPTOR
DESCRIPTOR.message_types_by_name['TxOut'] = _TXOUT
DESCRIPTOR.message_types_by_name['SignDescriptor'] = _SIGNDESCRIPTOR
DESCRIPTOR.message_types_by_name['SignReq'] = _SIGNREQ
DESCRIPTOR.message_types_by_name['SignResp'] = _SIGNRESP
DESCRIPTOR.message_types_by_name['InputScript'] = _INPUTSCRIPT
DESCRIPTOR.message_types_by_name['InputScriptResp'] = _INPUTSCRIPTRESP
DESCRIPTOR.message_types_by_name['SignMessageReq'] = _SIGNMESSAGEREQ
DESCRIPTOR.message_types_by_name['SignMessageResp'] = _SIGNMESSAGERESP
DESCRIPTOR.message_types_by_name['VerifyMessageReq'] = _VERIFYMESSAGEREQ
DESCRIPTOR.message_types_by_name['VerifyMessageResp'] = _VERIFYMESSAGERESP
DESCRIPTOR.message_types_by_name['SharedKeyRequest'] = _SHAREDKEYREQUEST
DESCRIPTOR.message_types_by_name['SharedKeyResponse'] = _SHAREDKEYRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

KeyLocator = _reflection.GeneratedProtocolMessageType('KeyLocator', (_message.Message,), {
  'DESCRIPTOR' : _KEYLOCATOR,
  '__module__' : 'signer_pb2'
  # @@protoc_insertion_point(class_scope:signrpc.KeyLocator)
  })
_sym_db.RegisterMessage(KeyLocator)

KeyDescriptor = _reflection.GeneratedProtocolMessageType('KeyDescriptor', (_message.Message,), {
  'DESCRIPTOR' : _KEYDESCRIPTOR,
  '__module__' : 'signer_pb2'
  # @@protoc_insertion_point(class_scope:signrpc.KeyDescriptor)
  })
_sym_db.RegisterMessage(KeyDescriptor)

TxOut = _reflection.GeneratedProtocolMessageType('TxOut', (_message.Message,), {
  'DESCRIPTOR' : _TXOUT,
  '__module__' : 'signer_pb2'
  # @@protoc_insertion_point(class_scope:signrpc.TxOut)
  })
_sym_db.RegisterMessage(TxOut)

SignDescriptor = _reflection.GeneratedProtocolMessageType('SignDescriptor', (_message.Message,), {
  'DESCRIPTOR' : _SIGNDESCRIPTOR,
  '__module__' : 'signer_pb2'
  # @@protoc_insertion_point(class_scope:signrpc.SignDescriptor)
  })
_sym_db.RegisterMessage(SignDescriptor)

SignReq = _reflection.GeneratedProtocolMessageType('SignReq', (_message.Message,), {
  'DESCRIPTOR' : _SIGNREQ,
  '__module__' : 'signer_pb2'
  # @@protoc_insertion_point(class_scope:signrpc.SignReq)
  })
_sym_db.RegisterMessage(SignReq)

SignResp = _reflection.GeneratedProtocolMessageType('SignResp', (_message.Message,), {
  'DESCRIPTOR' : _SIGNRESP,
  '__module__' : 'signer_pb2'
  # @@protoc_insertion_point(class_scope:signrpc.SignResp)
  })
_sym_db.RegisterMessage(SignResp)

InputScript = _reflection.GeneratedProtocolMessageType('InputScript', (_message.Message,), {
  'DESCRIPTOR' : _INPUTSCRIPT,
  '__module__' : 'signer_pb2'
  # @@protoc_insertion_point(class_scope:signrpc.InputScript)
  })
_sym_db.RegisterMessage(InputScript)

InputScriptResp = _reflection.GeneratedProtocolMessageType('InputScriptResp', (_message.Message,), {
  'DESCRIPTOR' : _INPUTSCRIPTRESP,
  '__module__' : 'signer_pb2'
  # @@protoc_insertion_point(class_scope:signrpc.InputScriptResp)
  })
_sym_db.RegisterMessage(InputScriptResp)

SignMessageReq = _reflection.GeneratedProtocolMessageType('SignMessageReq', (_message.Message,), {
  'DESCRIPTOR' : _SIGNMESSAGEREQ,
  '__module__' : 'signer_pb2'
  # @@protoc_insertion_point(class_scope:signrpc.SignMessageReq)
  })
_sym_db.RegisterMessage(SignMessageReq)

SignMessageResp = _reflection.GeneratedProtocolMessageType('SignMessageResp', (_message.Message,), {
  'DESCRIPTOR' : _SIGNMESSAGERESP,
  '__module__' : 'signer_pb2'
  # @@protoc_insertion_point(class_scope:signrpc.SignMessageResp)
  })
_sym_db.RegisterMessage(SignMessageResp)

VerifyMessageReq = _reflection.GeneratedProtocolMessageType('VerifyMessageReq', (_message.Message,), {
  'DESCRIPTOR' : _VERIFYMESSAGEREQ,
  '__module__' : 'signer_pb2'
  # @@protoc_insertion_point(class_scope:signrpc.VerifyMessageReq)
  })
_sym_db.RegisterMessage(VerifyMessageReq)

VerifyMessageResp = _reflection.GeneratedProtocolMessageType('VerifyMessageResp', (_message.Message,), {
  'DESCRIPTOR' : _VERIFYMESSAGERESP,
  '__module__' : 'signer_pb2'
  # @@protoc_insertion_point(class_scope:signrpc.VerifyMessageResp)
  })
_sym_db.RegisterMessage(VerifyMessageResp)

SharedKeyRequest = _reflection.GeneratedProtocolMessageType('SharedKeyRequest', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDKEYREQUEST,
  '__module__' : 'signer_pb2'
  # @@protoc_insertion_point(class_scope:signrpc.SharedKeyRequest)
  })
_sym_db.RegisterMessage(SharedKeyRequest)

SharedKeyResponse = _reflection.GeneratedProtocolMessageType('SharedKeyResponse', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDKEYRESPONSE,
  '__module__' : 'signer_pb2'
  # @@protoc_insertion_point(class_scope:signrpc.SharedKeyResponse)
  })
_sym_db.RegisterMessage(SharedKeyResponse)


DESCRIPTOR._options = None
_SHAREDKEYREQUEST.fields_by_name['key_loc']._options = None

_SIGNER = _descriptor.ServiceDescriptor(
  name='Signer',
  full_name='signrpc.Signer',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=1048,
  serialized_end=1388,
  methods=[
  _descriptor.MethodDescriptor(
    name='SignOutputRaw',
    full_name='signrpc.Signer.SignOutputRaw',
    index=0,
    containing_service=None,
    input_type=_SIGNREQ,
    output_type=_SIGNRESP,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ComputeInputScript',
    full_name='signrpc.Signer.ComputeInputScript',
    index=1,
    containing_service=None,
    input_type=_SIGNREQ,
    output_type=_INPUTSCRIPTRESP,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SignMessage',
    full_name='signrpc.Signer.SignMessage',
    index=2,
    containing_service=None,
    input_type=_SIGNMESSAGEREQ,
    output_type=_SIGNMESSAGERESP,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='VerifyMessage',
    full_name='signrpc.Signer.VerifyMessage',
    index=3,
    containing_service=None,
    input_type=_VERIFYMESSAGEREQ,
    output_type=_VERIFYMESSAGERESP,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeriveSharedKey',
    full_name='signrpc.Signer.DeriveSharedKey',
    index=4,
    containing_service=None,
    input_type=_SHAREDKEYREQUEST,
    output_type=_SHAREDKEYRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_SIGNER)

DESCRIPTOR.services_by_name['Signer'] = _SIGNER

# @@protoc_insertion_point(module_scope)



class RouteWithTooSmallCapacity(Exception):
    pass


class PaymentFailure(Exception):
    pass


class DryRun(Exception):
    pass


class PaymentTimeOut(Exception):
    pass


class TooExpensive(Exception):
    pass


class RebalanceFailure(Exception):
    pass


class NoRebalanceCandidates(RebalanceFailure):
    pass


class NotEconomic(RebalanceFailure):
    pass


class RebalancingTrialsExhausted(RebalanceFailure):
    pass


class PaymentError(Exception):
    pass


class PolicyError(PaymentError):
    pass


class OurNodeFailure(PaymentError):
    pass


class NoRoute(Exception):
    pass


class DuplicateRoute(NoRoute):
    pass


class RPCError(Exception):
    pass


class RouterRPCError(RPCError):
    pass


class TemporaryChannelFailure(PaymentFailure):
    def __init__(self, payment):
        self.payment = payment
        super().__init__()


class TemporaryNodeFailure(PaymentFailure):
    def __init__(self, payment):
        self.payment = payment
        super().__init__()


class UnknownNextPeer(PaymentFailure):
    def __init__(self, payment):
        self.payment = payment
        super().__init__()


class FeeInsufficient(PaymentFailure):
    def __init__(self, payment):
        self.payment = payment
        super().__init__()


class IncorrectCLTVExpiry(PaymentFailure):
    def __init__(self, payment):
        self.payment = payment
        super().__init__()


class ChannelDisabled(PaymentFailure):
    def __init__(self, payment):
        self.payment = payment
        super().__init__()



"""
Module for opening lightning channels in a batched way.
"""
from math import ceil
import logging
from typing import TYPE_CHECKING, List, Optional, Tuple

from lndmanage.lib.data_types import UTXO, AddressType
from lndmanage import settings

if TYPE_CHECKING:
    from lndmanage.lib.node import LndNode

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# transaction component sizes
# parameters via github.com/virtu/libtxsize
# ./libtxsize-cli.py -i P2SH-P2WPKH,P2WPKH -o P2WSH-2-of-2-MULTISIG,P2WPKH

# version, locktime, nins, nouts
TRANSACTION_OVERHEAD_VBYTES = 8 + 1 + 1
# inputs
P2WPKH_INPUT_VBYTES = 41
P2SH_P2WPKH_INPUT_VBYTES = 64
# outputs
P2WSH_OUTPUT_VBYTES = 43
P2WPKH_OUTPUT_VBYTES = 31
# witness
P2SH_P2WPKH_INPUT_WITNESS_WEIGHT = 109
P2WPKH_INPUT_WITNESS_WEIGHT = 109
MARKER_FLAG_WEIGHT = 2
WITNESS_NUM_INPUT_WEIGHT = 1
WITNESS_SCALE = 4

ANCHOR_RESERVE = 10 * 10000
MAX_ANCHOR_RESERVE = 5 * ANCHOR_RESERVE
MIN_CHANNEL_SIZE = 500000
WUMBO_LIMIT = 16777215

def calculate_fees(sat_per_vbyte: int, num_p2wkh_inputs: int, num_np2wkh_inputs,
                   num_channels: int, has_change=False) -> int:
    """Calculates the size (in vbytes) of a transaction and determines the fee."""
    # see also https://github.com/btcsuite/btcwallet/tree/master/wallet/txsizes/size.go
    # for lnd fee calculation
    size_vbytes = 0
    size_vbytes += TRANSACTION_OVERHEAD_VBYTES
    size_vbytes += P2WPKH_INPUT_VBYTES * num_p2wkh_inputs
    size_vbytes += P2SH_P2WPKH_INPUT_VBYTES * num_np2wkh_inputs
    size_vbytes += P2WSH_OUTPUT_VBYTES * num_channels

    # due to a bug in lnd fee estimation, we always add a change output here
    size_vbytes += P2WPKH_OUTPUT_VBYTES

    if has_change:
        size_vbytes += P2WPKH_OUTPUT_VBYTES

    # add witness data
    witness_weight = 0
    witness_weight += MARKER_FLAG_WEIGHT
    witness_weight += WITNESS_NUM_INPUT_WEIGHT
    witness_weight += P2WPKH_INPUT_WITNESS_WEIGHT * num_p2wkh_inputs
    witness_weight += P2SH_P2WPKH_INPUT_WITNESS_WEIGHT * num_np2wkh_inputs

    # for the total vbytes, discount witness data
    size_vbytes = size_vbytes + (witness_weight + 3) / WITNESS_SCALE

    logger.debug(
        f"    Transaction size calculation for "
        f"{num_p2wkh_inputs} (p2wkh) + {num_np2wkh_inputs} (np2wkh) inputs, {num_channels}"
        f"channels, change {has_change}: {int(size_vbytes)} vbytes"
    )
    # take an overestimation, let LND reduce fees (as LND's fee estimation is buggy)
    return int(size_vbytes * sat_per_vbyte)


def count_input_types(utxos: List[UTXO]) -> Tuple[int, int]:
    """Count the number of p2wkh and np2wkh inputs in the list of UTXOs."""
    num_p2wkh = 0
    num_npw2kh = 0
    for utxo in utxos:
        if utxo.transaction_type == AddressType.WITNESS_PUBKEY_HASH:
            num_p2wkh += 1
        elif utxo.transaction_type == AddressType.NESTED_PUBKEY_HASH:
            num_npw2kh += 1
    return num_p2wkh, num_npw2kh


def provide_coins(utxos: List[UTXO], total_amount_requested: Optional[int],
                  spend_all_utxos: bool, sat_per_vbyte: int, num_channels: int,
                  anchor_reserve: int) -> Tuple[List[UTXO], int, int, int]:
    """Selects coins from a list of UTXOs to spend a total amount.

    A total amount of None indicates that we want to spend all uxtos.

    If spend_all_utxos is true, all utxos will be included as the inputs, and
    change is accordingly created.
    The coin selection handles the reservation of anchor funds.

    :returns
    included UTXOS,
    total budget spendable in sat,
    change amount in sat,
    fee amount in sat
    """
    available_utxos = sorted(utxos, reverse=True)  # decreasing order

    if not total_amount_requested:  # try to spend the full UTXO set
        create_anchor_output = bool(anchor_reserve)
        # Here we try to spend all funds. If we should reserve some funds for
        # anchor outputs, we try to take some already existent small UTXOs. If
        # that's not possible, we need to create a change output.
        anchor_utxos = []
        if create_anchor_output:
            # Search for viable anchor UTXOs beginning from the smallest UTXO.
            for utxo in reversed(available_utxos):
                anchor_utxos.append(utxo)
                if anchor_reserve <= sum(utxo.amount_sat for utxo in anchor_utxos) <= MAX_ANCHOR_RESERVE:
                    logger.info("    Reserving UTXOs for anchors:")
                    for anchor_utxo in anchor_utxos:
                        logger.info(f"    {str(anchor_utxo)}")
                        available_utxos.remove(anchor_utxo)
                    create_anchor_output = False
                    break
        num_p2wkh, num_np2wkh = count_input_types(available_utxos)
        fee = calculate_fees(
            sat_per_vbyte,
            num_p2wkh_inputs=num_p2wkh,
            num_np2wkh_inputs=num_np2wkh,
            num_channels=num_channels,
            has_change=create_anchor_output
        )
        anchor_change = anchor_reserve if create_anchor_output else 0
        budget = sum(utxo.amount_sat for utxo in available_utxos) \
            - anchor_change - fee
        if budget < 0:
            raise ValueError(f"Not enough funds for channel opening.")
        change = anchor_change
        return available_utxos, budget, change, fee

    else:  # We try to spend a smaller amount than the given UTXO set.
        # We need to include change either because funds are left over
        # or we need to keep the anchor reserve.
        create_change = True
        if spend_all_utxos:  # user wants to batch all inputs together
            num_p2wkh, num_np2wkh = count_input_types(utxos)
            fee = calculate_fees(
                sat_per_vbyte,
                num_p2wkh_inputs=num_p2wkh,
                num_np2wkh_inputs=num_np2wkh,
                num_channels=num_channels,
                has_change=create_change
            )
            utxo_sum = sum(utxo.amount_sat for utxo in utxos)
            budget = min(total_amount_requested, utxo_sum - fee)
            if budget < 0:
                raise ValueError(f"Not enough funds for channel opening.")
            change = utxo_sum - budget - fee
            return utxos, budget, change, fee

        else:  # we need to do coin selection
            selected_utxos = []
            utxo_sum = 0
            for utxo in available_utxos:
                utxo_sum += utxo.amount_sat
                selected_utxos.append(utxo)
                num_p2wkh, num_np2wkh = count_input_types(selected_utxos)
                fee = calculate_fees(
                    sat_per_vbyte,
                    num_p2wkh_inputs=num_p2wkh,
                    num_np2wkh_inputs=num_np2wkh,
                    num_channels=num_channels,
                    has_change=True
                )
                if total_amount_requested + fee + anchor_reserve <= utxo_sum:
                    change = utxo_sum - total_amount_requested - fee
                    return selected_utxos, total_amount_requested, change, fee
    raise ValueError("Don't have enough funds.")


class ChannelOpener(object):
    """Opens multiple channels at once."""

    def __init__(self, node: 'LndNode'):
        self.node = node

    def _parse_pubkeys(self, pubkeys: str) -> List[str]:
        pubkeys = pubkeys.split(',')
        pubkeys = [c.strip() for c in pubkeys]
        return pubkeys

    def _pubkeys_to_bytes(self, pubkeys: List[str]) -> List[bytes]:
        return [bytes.fromhex(pubkey) for pubkey in pubkeys]

    @staticmethod
    def _parse_amounts(amounts: str) -> Optional[List[int]]:
        if amounts:
            amount_ints = amounts.split(',')
            amount_ints = [int(a) for a in amount_ints]
            return amount_ints
        else:
            return None

    @staticmethod
    def _parse_utxo_outpoints(utxos_str: str) -> Optional[List[UTXO]]:
        if utxos_str:
            utxo_strings = utxos_str.split(',')
            utxos = []
            for u in utxo_strings:
                try:
                    txid, output_index = u.split(':')
                except ValueError:
                    raise ValueError("utxo format is not of txid:index")
                utxos.append(
                    UTXO(txid=str(txid), output_index=int(output_index))
                )
            return utxos
        else:
            return None

    def open_channels(self, *, pubkeys: str, amounts: str = None,
                      utxos: Optional[str] = None, sat_per_vbyte=1,
                      total_amount: Optional[int] = None, reckless=False,
                      private=False):
        """
        Performs input checks on parameters and performs user interaction for
        batch channel opening.

        pubkeys: comma separated nodeid1,nodeid2,...
        amounts: comma separated amount1,amount2,...
        utxos: txid:output_idx,txid:output_idx,...
        sat_per_vbyte: onchain fee rate
        total_amount: if amounts are not specified, a total amount is
            distributed to all peers

        Steps:
        1. connect to nodes
        2. report about errors and demand actions (TODO)
        3. check provided utxos (if any)
        4. calculate budget without fees and anchor reserve
        5. open channels

        """
        # Possible improvements:
        # add close_addresses for upfront shutdown script
        pubkeys = self._parse_pubkeys(pubkeys)
        amounts = self._parse_amounts(amounts)

        if amounts and total_amount:
            raise ValueError("Specify either amounts or total amount.")

        if amounts and len(amounts) != len(pubkeys):
            raise ValueError("Number of amounts is not equal to number of"
                             "node pubkeys.")
        # 1. connect to nodes
        try:
            pubkeys_succeeded = self.node._connect_nodes(pubkeys)
        except ConnectionRefusedError:
            logger.info(">>> Could not connect to all nodes. Try to connect "
                        "manually via 'lncli connect', then rerun.")
            return

        assert len(pubkeys) == len(pubkeys_succeeded)
        pubkeys = self._pubkeys_to_bytes(pubkeys_succeeded)
        num_channels = len(pubkeys)

        # 2. report about errors
        # TODO: report about connection errors and how to proceed
        # maybe reduce number of peers, or suggest new list of peers that the
        # user can copy-paste

        # 3. check utxos
        wallet_utxos = self.node.get_utxos()
        wallet_balance = sum(utxo.amount_sat for utxo in wallet_utxos)
        user_provided_utxos = self._parse_utxo_outpoints(utxos)

        available_utxos = []
        if user_provided_utxos:
            for utxo in user_provided_utxos:
                if utxo in wallet_utxos:
                    # need to do index lookups here to also get additional info on the utxos
                    available_utxos.append(wallet_utxos[wallet_utxos.index(utxo)])
                else:
                    raise ValueError(f"UTXO {utxo} is not controlled by the wallet")
        else:
            available_utxos = wallet_utxos
        available_balance = sum(utxo.amount_sat for utxo in available_utxos)

        if not available_utxos:
            raise ValueError("no UTXOs available'")
        logger.info(">>> Available UTXOs:")
        for utxo in available_utxos:
            logger.debug(f"    {utxo.txid}:{utxo.output_index} {utxo.amount_sat} sat")

        if available_balance > wallet_balance - ANCHOR_RESERVE:
            # the user included maybe too many UTXOs that may spend beyond the anchor reserve
            logger.debug("    Need to potentially consider anchor outputs.")
            anchor_reserve = ANCHOR_RESERVE
        else:
            anchor_reserve = 0

        if (sum(utxo.amount_sat for utxo in available_utxos) - anchor_reserve) \
                / num_channels < MIN_CHANNEL_SIZE:
            raise ValueError(f"The total available funds are not enough to "
                             f"fund {num_channels} channels with minimal size of "
                             f"{MIN_CHANNEL_SIZE} sat. (Anchor reserve: "
                             f"{anchor_reserve} sat.)")

        # 4. calculate budget
        # Amounts and total_amount are mutually exclusive. Total amount of None
        # indicates a full spend.
        if amounts:
            total_amount = sum(amounts)
            if total_amount < 100:
                total_amount = None
        elif total_amount:
            amounts = [int(total_amount / num_channels) for _ in pubkeys]
        else:
            total_amount = None

        # If the total amount doesn't resepect the anchor reserve, make it a full spend.
        if total_amount and total_amount > available_balance - anchor_reserve:
            total_amount = None

        utxos, budget, change, fee = provide_coins(
            utxos=available_utxos,
            spend_all_utxos=bool(utxos),  # user provided some utxos
            total_amount_requested=total_amount,
            sat_per_vbyte=sat_per_vbyte,
            num_channels=num_channels,
            anchor_reserve=anchor_reserve,
        )
        logger.info(">>> Used UTXOs:")
        for utxo in utxos:
            logger.info(f"    {utxo.txid}:{utxo.output_index} {utxo.amount_sat} sat")
        logger.info(f">>> Channels will be {'private' if private else 'public'}.")

        def rescale(amounts: List[int], rescaled_total_amount: int) -> List[int]:
            """Rescales list of amounts to amounts with sum of
            rescaled_total_amount, fixing also rounding errors."""
            tot_amt = sum(amounts)
            amounts = [rescaled_total_amount * a // tot_amt for a in amounts]
            # handle rounding errors
            diff = rescaled_total_amount - sum(amounts)
            amounts[-1] += diff
            return amounts

        if not total_amount:
            if not amounts:
                # distribute the total budget equally over all channels
                amounts = [int(budget / num_channels) for _ in pubkeys]
            else:
                amounts = rescale(amounts, budget)

        logger.info(f"    Channel capacities: {amounts} sat ({sum(amounts)} sat total).")
        logger.info(f"    Fees: {fee} sat ({sat_per_vbyte} sat/vbyte).")
        logger.info(f"    Change: {change} sat.")

        # TODO: check connected node's features and allow wumbo channels
        # if own node supports it
        for amount in amounts:
            if amount > WUMBO_LIMIT:
                raise ValueError("Wumbo channels (capacity bigger than 16777215 sat) not yet supported.")

        # 5. open channels
        try:
            self.node.open_channels(
                pubkeys=pubkeys,
                amounts_sat=amounts,
                change_sat=change,
                utxos=utxos,
                reckless=reckless,
                private=private,
                sat_per_vbyte=sat_per_vbyte,
            )
        except Exception as e:
            logger.exception(e)
            pass



#!/usr/bin/env python
import argparse
import time
import os
import sys
import distutils.spawn as spawn
# readline has a desired side effect on keyword input of enabling history
import readline

from lndmanage.lib.exceptions import (
    DryRun,
    PaymentTimeOut,
    TooExpensive,
    RebalanceFailure,
    RebalancingTrialsExhausted,
)
from lndmanage.lib.fee_setting import FeeSetter, optimization_parameters
from lndmanage.lib.info import Info
from lndmanage.lib.listings import ListChannels, ListPeers
from lndmanage.lib.lncli import Lncli
from lndmanage.lib.node import LndNode
from lndmanage.lib.openchannels import ChannelOpener
from lndmanage.lib.rebalance import Rebalancer, DEFAULT_MAX_FEE_RATE, DEFAULT_AMOUNT_SAT
from lndmanage.lib.recommend_nodes import RecommendNodes
from lndmanage.lib.report import Report
from lndmanage import settings

import logging.config
logging.config.dictConfig(settings.logger_config)
logger = logging.getLogger()


def range_limited_float_type(unchecked_value):
    """
    Type function for argparse - a float within some predefined bounds

    :param: unchecked_value: float
    """
    try:
        value = float(unchecked_value)
    except ValueError:
        raise argparse.ArgumentTypeError("Must be a floating point number")
    if value < 1E-6 or value > 1:
        raise argparse.ArgumentTypeError(
            "Argument must be < " + str(1E-6) + " and > " + str(1))
    return value


def unbalanced_float(x):
    """
    Checks if the value is a valid unbalancedness between [-1 ... 1]
    """
    x = float(x)
    if x < -1.0 or x > 1.0:
        raise argparse.ArgumentTypeError(f"{x} not in range [-1.0, 1.0]")
    return x


class Parser(object):
    def __init__(self):

        # figure out if lncli is available and determine path of executable
        self.lncli_path = None

        self.check_for_lncli()

        # setup the command line parser
        self.parser = argparse.ArgumentParser(
            prog='lndmanage.py',
            description='Lightning network daemon channel management tool.')
        self.parser.add_argument(
            '--loglevel', default='INFO', choices=['INFO', 'DEBUG'])
        subparsers = self.parser.add_subparsers(dest='cmd')

        self.parser_status = subparsers.add_parser(
            'status', help='display node status',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

        # cmd: listchannels
        self.parser_listchannels = subparsers.add_parser(
            'listchannels',
            help='lists channels with extended information '
                 '[see also subcommands with -h]',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        listchannels_subparsers = self.parser_listchannels.add_subparsers(
            dest='subcmd')

        # subcmd: listchannels rebalance
        parser_listchannels_rebalance = listchannels_subparsers.add_parser(
            'rebalance', help='displays unbalanced channels')
        parser_listchannels_rebalance.add_argument(
            '--unbalancedness', type=float,
            default=settings.UNBALANCED_CHANNEL,
            help='Unbalancedness is a way to express how balanced a '
                 'channel is, a value between [-1, 1] (a perfectly balanced '
                 'channel has a value of 0). The flag excludes channels with '
                 'an absolute unbalancedness smaller than UNBALANCEDNESS.')
        parser_listchannels_rebalance.add_argument(
            '--sort-by', default='rev_ub', type=str,
            help='sort by column (look at description)')

        # subcmd: listchannels inactive
        parser_listchannels_inactive = listchannels_subparsers.add_parser(
            'inactive', help="displays inactive channels")
        parser_listchannels_inactive.add_argument(
            '--sort-by', default='lupp', type=str,
            help='sort by column (look at description)')

        # subcmd: listchannels forwardings
        parser_listchannels_forwardings = listchannels_subparsers.add_parser(
            'forwardings',
            help="displays channels with forwarding information")
        parser_listchannels_forwardings.add_argument(
            '--from-days-ago', default=30, type=int,
            help='time interval start (days ago)')
        parser_listchannels_forwardings.add_argument(
            '--to-days-ago', default=0, type=int,
            help='time interval end (days ago)')
        parser_listchannels_forwardings.add_argument(
            '--sort-by', default='fo/w', type=str,
            help='sort by column (look at description)')

        # subcmd: listchannels hygiene
        parser_listchannels_hygiene = listchannels_subparsers.add_parser(
            'hygiene',
            help="displays channels with information for channel closing")
        parser_listchannels_hygiene.add_argument(
            '--from-days-ago', default=60, type=int,
            help='time interval start (days ago)')
        parser_listchannels_hygiene.add_argument(
            '--sort-by', default='rev_nfwd/a', type=str,
            help='sort by column (look at description)')

        # cmd: listpeers
        self.parser_listpeers = subparsers.add_parser(
            'listpeers',
            help='lists peers with extended information',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        self.parser_listpeers.add_argument(
            '--from-days-ago', default=60, type=int,
            help='time interval start (days ago)')
        self.parser_listpeers.add_argument(
            '--sort-by', default='fio', type=str,
            help='sort by column (look at description)')
        listpeers_subparsers = self.parser_listpeers.add_subparsers(
            dest='subcmd')

        # cmd: listpeers in
        listpeers_subparsers.add_parser(
            'in',
            help="displays peers sorted by inward traffic")

        # cmd: listpeers out
        listpeers_subparsers.add_parser(
            'out',
            help="displays peers sorted by outward traffic")

        # cmd: rebalance
        self.parser_rebalance = subparsers.add_parser(
            'rebalance', help='rebalance a channel',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        self.parser_rebalance.add_argument('node_channel', type=str,
                                           help='node id or channel id')
        self.parser_rebalance.add_argument(
            '--max-fee-sat', type=int, default=None,
            help='Sets the maximal fees in satoshis to be paid.')
        self.parser_rebalance.add_argument(
            '--amount', type=int, default=None,
            help='Specifies the increase in local balance in sat. The amount can be'
                 f'negative to decrease the local balance. Default: {DEFAULT_AMOUNT_SAT} sat.')
        self.parser_rebalance.add_argument(
            '--max-fee-rate', type=range_limited_float_type, default=DEFAULT_MAX_FEE_RATE,
            help='Sets the maximal effective fee rate to be paid.'
                 ' The effective fee rate is defined by '
                 '(base_fee + amt * fee_rate) / amt.')
        self.parser_rebalance.add_argument(
            '--reckless', help='Execute action in the network.',
            action='store_true')
        self.parser_rebalance.add_argument(
            '--force',
            help=f"Allow rebalances that are uneconomic.",
            action='store_true')
        self.parser_rebalance.add_argument(
            '--target', help=f'The unbalancedness target is between [-1, 1]. '
            f'A target of -1 leads to a maximal local balance, a target of 0 '
            f'to a 50:50 balanced channel and a target of 1 to a maximal '
            f'remote balance. Default is a target of 0.',
            type=unbalanced_float, default=None)

        # cmd: recommend-nodes
        self.parser_recommend_nodes = subparsers.add_parser(
            'recommend-nodes',
            help='recommends nodes [see also subcommands with -h]',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        self.parser_recommend_nodes.add_argument(
            '--show-connected', action='store_true', default=False,
            help='specifies if already connected nodes should be '
                 'removed from list')
        self.parser_recommend_nodes.add_argument(
            '--show-addresses', action='store_true', default=False,
            help='specifies if node addresses should be shown')
        parser_recommend_nodes_subparsers = \
            self.parser_recommend_nodes.add_subparsers(
                dest='subcmd')

        # TODO: put global options to the
        #  parent parser (e.g. number of nodes, sort-by flag)

        # subcmd: recommend-nodes good-old
        parser_recommend_nodes_good_old = \
            parser_recommend_nodes_subparsers.add_parser(
                'good-old',
                help='nodes with previous good relationship (channels)',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser_recommend_nodes_good_old.add_argument(
            '--nnodes', default=20, type=int,
            help='sets the number of nodes displayed')
        parser_recommend_nodes_good_old.add_argument(
            '--sort-by', default='tot', type=str,
            help="sort by column [abbreviation, e.g. 'tot']")

        # subcmd: recommend-nodes flow-analysis
        parser_recommend_nodes_flow_analysis = \
            parser_recommend_nodes_subparsers.add_parser(
                'flow-analysis', help='nodes from a flow analysis',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser_recommend_nodes_flow_analysis.add_argument(
            '--nnodes', default=20, type=int,
            help='sets the number of nodes displayed')
        parser_recommend_nodes_flow_analysis.add_argument(
            '--forwarding-events', default=200, type=int,
            help='sets the number of forwarding events in the flow analysis')
        parser_recommend_nodes_flow_analysis.add_argument(
            '--inwards', action='store_true',
            help='if True, inward-flowing nodes are displayed '
                 'instead of outward-flowing nodes')
        parser_recommend_nodes_flow_analysis.add_argument(
            '--sort-by', default='weight', type=str,
            help="sort by column [abbreviation, e.g. 'nchan']")

        # subcmd: recommend-nodes external_source
        parser_recommend_nodes_external_source = \
            parser_recommend_nodes_subparsers.add_parser(
                'external-source',
                help='nodes from a given file/url',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser_recommend_nodes_external_source.add_argument(
            '--nnodes', default=20, type=int,
            help='sets the number of nodes displayed')
        parser_recommend_nodes_external_source.add_argument(
            '--source', type=str,
            default='https://github.com/lightningnetworkstores/'
                    'lightningnetworkstores.github.io/raw/master/sites.json',
            help='url/file to be analyzed')
        parser_recommend_nodes_external_source.add_argument(
            '--distributing-nodes', action='store_true',
            help='if True, distributing nodes are '
                 'displayed instead of the bare nodes')
        parser_recommend_nodes_external_source.add_argument(
            '--sort-by', default='cpc', type=str,
            help="sort by column [abbreviation, e.g. 'nchan']")

        # subcmd: recommend-nodes channel-openings
        parser_recommend_nodes_channel_openings = \
            parser_recommend_nodes_subparsers.add_parser(
                'channel-openings',
                help='nodes from recent channel openings',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser_recommend_nodes_channel_openings.add_argument(
            '--nnodes', default=20, type=int,
            help='sets the number of nodes displayed')
        parser_recommend_nodes_channel_openings.add_argument(
            '--from-days-ago', type=int,
            default=30,
            help='channel openings starting from a time frame days ago')
        parser_recommend_nodes_channel_openings.add_argument(
            '--sort-by', default='msteady', type=str,
            help="sort by column [abbreviation, e.g. 'nchan']")

        # subcmd: recommend-nodes second-neighbors
        parser_recommend_nodes_second_neighbors = \
            parser_recommend_nodes_subparsers.add_parser(
                'second-neighbors',
                help='nodes from network analysis giving most '
                     'second neighbors',
                description="This command recommends nodes for getting more "
                            "second neighbors. "
                            "This is achieved by checking how many second "
                            "neighbors would be added if one would connect to "
                            "the suggested node. A channel to the node "
                            "should get your node closer to "
                            "more other nodes.",
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser_recommend_nodes_second_neighbors.add_argument(
            '--nnodes', default=20, type=int,
            help='sets the number of nodes displayed')
        parser_recommend_nodes_second_neighbors.add_argument(
            '--sort-by', default='sec', type=str,
            help="sort by column [abbreviation, e.g. 'sec']")

        # cmd: report
        parser_report = subparsers.add_parser(
            'report',
            help="displays reports of activity on the node")
        parser_report.add_argument(
            '--from-days-ago', default=1, type=int,
            help='time interval start (days ago)')
        parser_report.add_argument(
            '--to-days-ago', default=0, type=int,
            help='time interval end (days ago)')

        # cmd: info
        parser_info = subparsers.add_parser(
            'info',
            help='displays info on channels and nodes')
        parser_info.add_argument(
            'info_string', type=str,
            help='info string can represent a node public key or a channel id')

        # cmd: lncli
        if self.lncli_path:
            parser_lncli = subparsers.add_parser(
            'lncli',
            help='execute lncli')

        # cmd: openchannels
        self.parser_openchannels = subparsers.add_parser(
            'openchannels',
            help='opens multiple channels with UTXO control',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        self.parser_openchannels.add_argument(
            '--utxos',
            type=str,
            help='Comma-separated list of utxos of format '
                 'txid:output_index,txid:output_index which belong to the LND '
                 'wallet (see lncli listunspent).')
        self.parser_openchannels.add_argument(
            '--amounts',
            type=str,
            help='Comma-separated list of channel amounts in sat 1000000,2000000,... '
                 'A list of amounts with sum < 100 denotes relative amounts that are rescaled '
                 'to the sum of all UTXOs provided, such that no change is created. '
                 'If the amounts provided are larger than the funds available, '
                 'they will automatically be reduced proportionally such that no change is created. '
                 'Change is created if anchor reserve funds would be depleted.')
        self.parser_openchannels.add_argument(
            '--total-amount',
            type=int,
            help='Total amount in sat to open channels.'
                 'Amounts and total amount flags are mutually exclusive. '
                 'If none of the two amounts are given, all available UTXOs '
                 'will be spent.')
        self.parser_openchannels.add_argument(
            '--sat-per-vbyte',
            type=int,
            default=1,
            help='The fee rate in sat per vbyte that will be targeted.')
        self.parser_openchannels.add_argument(
            '--reckless',
            help='If this flag is provided, the user will not be asked '
                 'to agree to channel opening.', action='store_true')
        self.parser_openchannels.add_argument(
            '--private', help='The channels will not be announced to the network.',
            action='store_true')
        self.parser_openchannels.add_argument(
            'pubkeys',
            type=str,
            help='Comma-separated list of node pubkeys.')

        # cmd: update-fees
        self.parser_update_fees = subparsers.add_parser(
            'update-fees',
            description='Periodically running this command increases/decreases'
                        'the fees on all channels by adapting them according to '
                        'the forwarding demand in the last interval, which can '
                        'be set by the parameter --from-days-ago. The fee optimization '
                        'tries to keep a liquidity buffer for excess-demand times.'
                        'Channels can be excluded via the config section'
                        'excluded-channels-fee-opt.\n'
                        'The command will prompt the fees it would set after a yes/no question.'
                        "\n\n**Don't run this command too frequently (only once a week), "
                        "otherwise you put strain on the network and the new"
                        "fee policies won't reach end points like mobile phones"
                        "and you will route less.**",
            help='optimize the fees on your channels to increase revenue and to automatically rebalance',
            formatter_class=argparse.RawDescriptionHelpFormatter)
        self.parser_update_fees.add_argument(
            '--cltv', type=int, default=optimization_parameters['cltv'],
            help='CLTV time delta.')
        self.parser_update_fees.add_argument(
            '--min-base-fee-msat', type=int,
            default=optimization_parameters['min_base_fee'],
            help='The base fee cannot go lower than this.')
        self.parser_update_fees.add_argument(
            '--max-base-fee-msat', type=int,
            default=optimization_parameters['max_base_fee'],
            help='The base fee cannot go higher than this.')
        self.parser_update_fees.add_argument(
            '--min-fee-rate', type=float,
            default=optimization_parameters['min_fee_rate'],
            help='The fee rate cannot go lower than this.')
        self.parser_update_fees.add_argument(
            '--max-fee-rate', type=float,
            default=optimization_parameters['max_fee_rate'],
            help='The fee rate cannot go higher than this.'
            'Half of this value is also used for initialization.'
        )
        self.parser_update_fees.add_argument(
            '--init', action='store_true',
            help='If set, uses half of max-fee-rate for fee rates.')
        self.parser_update_fees.add_argument(
            '--from-days-ago', type=int,
            default=7,
            help='Sets the number of days over which forwarding action is taken'
                 'into account. This value should coincide with the fee '
                 'update interval.')
        self.parser_update_fees.add_argument(
            '--target-forwarding-amount-sat', type=int,
            default=optimization_parameters['r_t'],
            help='The target for how much a channel should route per day.'
                 'The value of this parameter will influence how much you earn in forwarding'
                 'fees. If you set it too low, no forwardings will happen. If you set it too'
                 'high, you sell your liquidity too cheaply.'
                 'A good value could be half the amount you route in your most-income channel per day.')
        self.parser_update_fees.add_argument(
            '--reckless',
            help='Update the fees without asking the user explicitly.',
            action='store_true')

    def check_for_lncli(self):
        """
        Looks for lncli in PATH or in LNDMANAGE_HOME folder. Sets self.lncli_path.
        Executable in LNDMANAGE_HOME is prioritized.
        """
        lncli_candidate = os.path.join(settings.home_dir, 'lncli')

        # look in lndmanage home folder after lncli
        if os.access(lncli_candidate, os.X_OK):
            self.lncli_path = lncli_candidate
        # look in PATH
        else:
            path = spawn.find_executable('lncli')
            self.lncli_path = path

    def parse_arguments(self):
        return self.parser.parse_args()

    def run_commands(self, node, args):
        # program execution
        if args.loglevel:
            # update the loglevel of the stdout handler to the user choice
            logger.handlers[0].setLevel(args.loglevel)

        if args.cmd == 'status':
            node.print_status()

        elif args.cmd == 'listchannels':
            listchannels = ListChannels(node)
            if not args.subcmd:
                listchannels.print_all_channels('rev_alias')
            if args.subcmd == 'rebalance':
                listchannels.print_channels_unbalanced(
                    args.unbalancedness, sort_string=args.sort_by)
            elif args.subcmd == 'inactive':
                listchannels.print_channels_inactive(
                    sort_string=args.sort_by)
            elif args.subcmd == 'forwardings':
                # convert time interval into unix timestamp
                time_from = time.time() - args.from_days_ago * 24 * 60 * 60
                time_to = time.time() - args.to_days_ago * 24 * 60 * 60
                logger.info(
                    f"Forwardings from {args.from_days_ago} days ago"
                    f" to {args.to_days_ago} days ago are included.")
                listchannels.print_channels_forwardings(
                    time_interval_start=time_from, time_interval_end=time_to,
                    sort_string=args.sort_by)
            elif args.subcmd == 'hygiene':
                time_from = time.time() - args.from_days_ago * 24 * 60 * 60
                logger.info(f"Channel hygiene stats is over last "
                            f"{args.from_days_ago} days.")
                listchannels.print_channels_hygiene(
                    time_interval_start=time_from, sort_string=args.sort_by)

        elif args.cmd == 'listpeers':
            listpeers = ListPeers(node)
            time_from = time.time() - args.from_days_ago * 24 * 60 * 60
            time_to = time.time()
            logger.info(
                f"Forwardings from {args.from_days_ago} days ago"
                f" to now are included.")
            if not args.subcmd:
                listpeers.print_all_nodes(
                    time_interval_start=time_from,
                    time_interval_end=time_to,
                    sort_string=args.sort_by,
                )
            elif args.subcmd == 'in':
                listpeers.print_all_nodes(
                    time_interval_start=time_from,
                    time_interval_end=time_to,
                    sort_string='in',
                )
            elif args.subcmd == 'out':
                listpeers.print_all_nodes(
                    time_interval_start=time_from,
                    time_interval_end=time_to,
                    sort_string='out',
                )

        elif args.cmd == 'rebalance':
            if args.target:
                logger.warning("Warning: Target is set, this is still an "
                               "experimental feature.")
            rebalancer = Rebalancer(node, args.max_fee_rate, args.max_fee_sat, args.force)
            try:
                rebalancer.rebalance(
                    args.node_channel,
                    dry=not args.reckless,
                    target=args.target,
                    amount_sat=args.amount
                )
            except ValueError as e:
                logger.error(e)
            except TooExpensive as e:
                logger.error(f"Too expensive: {e}")
            except RebalanceFailure as e:
                logger.error(f"Rebalance failure: {e}")
            except KeyboardInterrupt:
                pass

        elif args.cmd == 'recommend-nodes':
            if not args.subcmd:
                self.parser_recommend_nodes.print_help()
                return 0

            recommend_nodes = RecommendNodes(
                node, show_connected=args.show_connected,
                show_addresses=args.show_addresses)

            if args.subcmd == 'good-old':
                recommend_nodes.print_good_old(number_of_nodes=args.nnodes,
                                               sort_by=args.sort_by)
            elif args.subcmd == 'flow-analysis':
                recommend_nodes.print_flow_analysis(
                    out_direction=(not args.inwards),
                    number_of_nodes=args.nnodes,
                    forwarding_events=args.forwarding_events,
                    sort_by=args.sort_by)
            elif args.subcmd == 'external-source':
                recommend_nodes.print_external_source(
                    args.source, distributing_nodes=args.distributing_nodes,
                    number_of_nodes=args.nnodes, sort_by=args.sort_by)
            elif args.subcmd == 'channel-openings':
                recommend_nodes.print_channel_openings(
                    from_days_ago=args.from_days_ago,
                    number_of_nodes=args.nnodes, sort_by=args.sort_by)
            elif args.subcmd == 'second-neighbors':
                recommend_nodes.print_second_neighbors(
                    number_of_nodes=args.nnodes, sort_by=args.sort_by)

        elif args.cmd == 'report':
            time_from = time.time() - args.from_days_ago * 24 * 60 * 60
            time_to = time.time() - args.to_days_ago * 24 * 60 * 60
            report = Report(node, time_from, time_to)
            report.report()

        elif args.cmd == 'info':
            info = Info(node)
            info.parse_and_print(args.info_string)

        elif args.cmd == 'openchannels':
            channel_opener = ChannelOpener(node)
            try:
                channel_opener.open_channels(
                    pubkeys=args.pubkeys,
                    amounts=args.amounts,
                    utxos=args.utxos,
                    sat_per_vbyte=args.sat_per_vbyte,
                    total_amount=args.total_amount,
                    reckless=args.reckless,
                    private=args.private,
                )
            except Exception as e:
                logger.info(e)
        elif args.cmd == 'update-fees':
            # overwrite default optimization parameters
            optimization_parameters['cltv'] = args.cltv
            optimization_parameters['min_base_fee'] = args.min_base_fee_msat
            optimization_parameters['max_base_fee'] = args.max_base_fee_msat
            optimization_parameters['min_fee_rate'] = args.min_fee_rate
            optimization_parameters['max_fee_rate'] = args.max_fee_rate
            optimization_parameters['r_t'] = args.target_forwarding_amount_sat

            feesetter = FeeSetter(
                node,
                from_days_ago=args.from_days_ago,
                parameters=optimization_parameters
            )

            feesetter.set_fees(
                init=args.init,
                reckless=args.reckless
            )


def main():
    parser = Parser()

    # config.ini is expected to be in home/.lndmanage directory
    config_file = os.path.join(settings.home_dir, 'config.ini')

    # if lndmanage is run with arguments, run once
    if len(sys.argv) > 1:
        # take arguments from sys.argv
        args = parser.parse_arguments()
        node = LndNode(config_file=config_file)
        parser.run_commands(node, args)

    # otherwise enter an interactive mode
    else:
        history_file = os.path.join(settings.home_dir, "command_history")
        try:
            readline.read_history_file(history_file)
        except FileNotFoundError:
            # history will be written later
            pass

        logger.info("Running in interactive mode. "
                    "You can type 'help' or 'exit'.")
        node = LndNode(config_file=config_file)

        if parser.lncli_path:
            logger.info("Enabled lncli: using " + parser.lncli_path)

        while True:
            try:
                user_input = input("$ lndmanage ")
            except KeyboardInterrupt:
                logger.info("")
                continue
            except EOFError:
                readline.write_history_file(history_file)
                logger.info("exit")
                return 0

            if not user_input or user_input in ['help', '-h', '--help']:
                parser.parser.print_help()
                continue
            elif user_input == 'exit':
                readline.write_history_file(history_file)
                return 0

            args_list = user_input.split(" ")

            # lncli execution
            if args_list[0] == 'lncli':
                if parser.lncli_path:
                    lncli = Lncli(parser.lncli_path, config_file)
                    lncli.lncli(args_list[1:])
                    continue
                else:
                    logger.info("lncli not enabled, put lncli in PATH or in ~/.lndmanage")
                    continue
            try:
                # need to run with parse_known_args to get an exception
                args = parser.parser.parse_args(args_list)
                parser.run_commands(node, args)
            except SystemExit:
                # argparse may raise SystemExit on incorrect user input,
                # which is a graceful exit. The user gets the standard output
                # from argparse of what went wrong.
                continue


if __name__ == '__main__':
    main()



import binascii
import codecs
from collections import OrderedDict, defaultdict
import datetime
import os
import time
from typing import List, TYPE_CHECKING, Optional, Dict

import grpc
from grpc._channel import _Rendezvous
from grpc._channel import _InactiveRpcError
from google.protobuf.json_format import MessageToDict

import lndmanage.grpc_compiled.lightning_pb2 as lnd
import lndmanage.grpc_compiled.lightning_pb2_grpc as lndrpc
import lndmanage.grpc_compiled.router_pb2 as lndrouter
import lndmanage.grpc_compiled.router_pb2_grpc as lndrouterrpc
import lndmanage.grpc_compiled.walletkit_pb2 as lndwalletkit
import lndmanage.grpc_compiled.walletkit_pb2_grpc as lndwalletkitrpc

from lndmanage.lib.network import Network
from lndmanage.lib.exceptions import PaymentTimeOut, NoRoute, OurNodeFailure
from lndmanage.lib import exceptions
from lndmanage.lib.ln_utilities import (
    extract_short_channel_id_from_string,
    convert_short_channel_id_to_channel_id,
    convert_channel_id_to_short_channel_id,
    local_balance_to_unbalancedness
)
from lndmanage.lib.psbt import extract_psbt_inputs_outputs
from lndmanage.lib.data_types import UTXO, AddressType
from lndmanage.lib.user import yes_no_question
from lndmanage.lib.utilities import convert_dictionary_number_strings_to_ints
from lndmanage import settings

if TYPE_CHECKING:
    from lndmanage.lib.routing import Route

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

NUM_MAX_FORWARDING_EVENTS = 100000
OPEN_EXPIRY_TIME_MINUTES = 8


class Node(object):
    """Bare node object with attributes."""
    def __init__(self):
        self.alias = ''
        self.pub_key = ''
        self.total_capacity = 0
        self.total_local_balance = 0
        self.total_remote_balance = 0
        self.total_channels = 0
        self.num_active_channels = 0
        self.num_peers = 0
        self.total_satoshis_received = 0
        self.total_satoshis_sent = 0
        self.total_private_channels = 0
        self.total_active_channels = 0
        self.blockheight = 0


class LndNode(Node):
    """Implements the node interface for LND."""
    def __init__(self, config_file: Optional[str] = None,
                 lnd_home: Optional[str] = None,
                 lnd_host: Optional[str] = None, regtest=False):
        """
        :param config_file: path to the config file
        :param lnd_home: path to lnd home folder
        :param lnd_host: lnd host of format "127.0.0.1:9735"
        :param regtest: if the node is representing a regtest node
        """
        super().__init__()
        if config_file:
            self.config_file = config_file
            self.config = settings.read_config(self.config_file)
        else:
            self.config_file = None
            self.config = None
        self.lnd_home = lnd_home
        self.lnd_host = lnd_host
        self.regtest = regtest

        self._rpc = None
        self._routerrpc = None
        self.connect_rpcs()

        self.set_info()
        self.network = Network(self)
        self.set_channel_summary()
        self.update_blockheight()

    def connect_rpcs(self):
        """
        Establishes a connection to lnd using the hostname, tls certificate,
        and admin macaroon defined in settings.
        """
        macaroons = True
        os.environ['GRPC_SSL_CIPHER_SUITES'] = \
            'ECDHE-RSA-AES128-GCM-SHA256:' \
            'ECDHE-RSA-AES128-SHA256:' \
            'ECDHE-RSA-AES256-SHA384:' \
            'ECDHE-RSA-AES256-GCM-SHA384:' \
            'ECDHE-ECDSA-AES128-GCM-SHA256:' \
            'ECDHE-ECDSA-AES128-SHA256:' \
            'ECDHE-ECDSA-AES256-SHA384:' \
            'ECDHE-ECDSA-AES256-GCM-SHA384'

        # if no lnd_home is given, then use the paths from the config,
        # else override them with default file paths in lnd_home
        if self.lnd_home is not None:
            cert_file = os.path.join(self.lnd_home, 'tls.cert')
            bitcoin_network = 'regtest' if self.regtest else 'mainnet'
            macaroon_file = os.path.join(
                self.lnd_home, 'data/chain/bitcoin/',
                bitcoin_network, 'admin.macaroon')
            if self.lnd_host is None:
                raise ValueError(
                    'if lnd_home is given, lnd_host must be given also')
            lnd_host = self.lnd_host
        else:
            cert_file = os.path.expanduser(self.config['network']['tls_cert_file'])
            macaroon_file = \
                os.path.expanduser(self.config['network']['admin_macaroon_file'])
            lnd_host = self.config['network']['lnd_grpc_host']

        cert = None
        try:
            with open(cert_file, 'rb') as f:
                cert = f.read()
        except FileNotFoundError:
            logger.error("tls.cert not found, please configure %s.",
                         self.config_file)
            exit(1)

        if macaroons:
            try:
                with open(macaroon_file, 'rb') as f:
                    macaroon_bytes = f.read()
                    macaroon = codecs.encode(macaroon_bytes, 'hex')
            except FileNotFoundError:
                logger.error("admin.macaroon not found, please configure %s.",
                             self.config_file)
                exit(1)

            def metadata_callback(context, callback):
                callback([('macaroon', macaroon)], None)

            cert_creds = grpc.ssl_channel_credentials(cert)
            auth_creds = grpc.metadata_call_credentials(metadata_callback)
            creds = grpc.composite_channel_credentials(cert_creds, auth_creds)

        else:
            creds = grpc.ssl_channel_credentials(cert)

        # necessary to circumvent standard size limitation
        channel = grpc.secure_channel(lnd_host, creds, options=[
            ('grpc.max_receive_message_length', 50 * 1024 * 1024)
        ])

        # establish connections to rpc servers
        self._rpc = lndrpc.LightningStub(channel)
        self._routerrpc = lndrouterrpc.RouterStub(channel)
        self._walletrpc = lndwalletkitrpc.WalletKitStub(channel)

    def update_blockheight(self):
        info = self._rpc.GetInfo(lnd.GetInfoRequest())
        self.blockheight = int(info.block_height)

    def get_channel_info(self, channel_id):
        channel = self._rpc.GetChanInfo(
            lnd.ChanInfoRequest(chan_id=channel_id))
        channel_dict = MessageToDict(
            channel, including_default_value_fields=True)
        channel_dict = convert_dictionary_number_strings_to_ints(channel_dict)
        return channel_dict

    @staticmethod
    def lnd_hops(hops) -> List[lnd.Hop]:
        return [lnd.Hop(**hop) for hop in hops]

    def _to_lnd_route(self, route: 'Route') -> lnd.Route:
        """
        Converts a cleartext route to an lnd route.
        """
        hops = self.lnd_hops(route.hops)
        lnd_route = lnd.Route(
            total_time_lock=route.total_time_lock,
            total_fees=route.total_fee_msat // 1000,
            total_amt=route.total_amt_msat // 1000,
            hops=hops,
            total_fees_msat=route.total_fee_msat,
            total_amt_msat=route.total_amt_msat
        )
        return lnd_route

    def get_invoice(self, amt_msat: int, memo: str) -> lnd.Invoice:
        """
        Creates a new invoice with amt_msat and memo.

        :param amt_msat: int
        :param memo: str
        :return: Hash of invoice preimage.
        """
        invoice = self._rpc.AddInvoice(lnd.Invoice(
            value=amt_msat // 1000, memo=memo))
        return invoice

    def get_rebalance_invoice(self, memo) -> lnd.Invoice:
        """
        Creates a zero amount invoice and gives back it's hash.

        :param memo: Comment for the invoice.
        :return: Hash of the invoice preimage.
        """
        invoice = self._rpc.AddInvoice(lnd.Invoice(value=0, memo=memo))
        return invoice

    def send_to_route(self, route: 'Route', payment_hash: bytes,
                      payment_address: bytes):
        """
        Takes bare route (list) and tries to send along it,
        trying to fulfill the invoice labeled by the given hash.

        :param route: (list) of :class:`lib.routes.Route`
        :param payment_hash: invoice identifier
        :return:
        """
        lnd_route = self._to_lnd_route(route)
        # set payment address for last hop
        lnd_route.hops[-1].tlv_payload = True
        lnd_route.hops[-1].mpp_record.payment_addr = payment_address
        lnd_route.hops[-1].mpp_record.total_amt_msat = lnd_route.hops[-1].amt_to_forward_msat
        # set payment hash
        request = lndrouter.SendToRouteRequest(
            route=lnd_route,
            payment_hash=payment_hash,
        )
        try:
            # timeout after 5 minutes
            payment = self._routerrpc.SendToRouteV2(request, timeout=5 * 60)
        except _Rendezvous:
            raise PaymentTimeOut
        if payment.HasField('failure'):
            failure = payment.failure  # type: lnd.Failure.FailureCode
            logger.debug(f"Routing failure: {failure}")
            if failure.failure_source_index == 0:
                raise OurNodeFailure("Not enough funds?")
            if failure.code == 12:
                raise exceptions.FeeInsufficient(payment)
            elif failure.code == 13:
                raise exceptions.IncorrectCLTVExpiry(payment)
            elif failure.code == 14:
                raise exceptions.ChannelDisabled(payment)
            elif failure.code == 15:
                raise exceptions.TemporaryChannelFailure(payment)
            elif failure.code == 18:
                raise exceptions.UnknownNextPeer(payment)
            elif failure.code == 19:
                raise exceptions.TemporaryNodeFailure(payment)
            else:
                logger.info(failure)
                raise Exception(f"Unknown error: code {failure.code}")

        return payment

    def get_raw_network_graph(self):
        try:
            graph = self._rpc.DescribeGraph(lnd.ChannelGraphRequest())
            return graph
        except _Rendezvous:
            logger.error(
                "Problem connecting to lnd. "
                "Either %s is not configured correctly or lnd is not running.",
                self.config_file)
            exit(1)

    def get_raw_info(self):
        """
        Returns specific information about this node.

        :return: node information
        """
        return self._rpc.GetInfo(lnd.GetInfoRequest())

    def set_info(self):
        """
        Fetches information about this node and computes total capacity,
        local and remote total balance, how many satoshis were sent and
        received, and some networking peer stats.
        """

        raw_info = self.get_raw_info()
        self.pub_key = raw_info.identity_pubkey
        self.alias = raw_info.alias
        self.num_active_channels = raw_info.num_active_channels
        self.num_peers = raw_info.num_peers

    def set_channel_summary(self):
        # TODO: remove the following code and implement an advanced status
        all_channels = self.get_open_channels(
            active_only=False, public_only=False)
        self.total_channels = len(all_channels)

        for k, c in all_channels.items():
            self.total_capacity += c['capacity']
            self.total_local_balance += c['local_balance']
            self.total_remote_balance += c['remote_balance']
            self.total_satoshis_received += c['total_satoshis_received']
            self.total_satoshis_sent += c['total_satoshis_sent']
            if c['active']:
                self.total_active_channels += 1
            if c['private']:
                self.total_private_channels += 1

    def get_open_channels(self, active_only=False, public_only=False) \
            -> Dict[int, Dict]:
        """
        Fetches information (fee settings of the counterparty, channel
        capacity, balancedness) about this node's open channels and saves
        it into the channels dict attribute.

        :param active_only: only take active channels into
                            account (off by default)
        :type active_only: bool
        :param public_only: only take public channels into
                            account (off by default)
        :type public_only: bool

        :return: dict of channels sorted by remote pubkey
        :rtype: OrderedDict

        """
        raw_channels = self._rpc.ListChannels(lnd.ListChannelsRequest(
            active_only=active_only, public_only=public_only))
        try:
            channels_data = raw_channels.ListFields()[0][1]
        except IndexError:
            # If there are no channels, return.
            return OrderedDict({})

        channels = OrderedDict()

        for c in channels_data:
            # calculate age from blockheight
            blockheight, _, _ = convert_channel_id_to_short_channel_id(
                c.chan_id)
            age_days = (self.blockheight - blockheight) * 10 / (60 * 24)
            try:
                sent_received_per_week = int(
                    (c.total_satoshis_sent +
                     c.total_satoshis_received) / (age_days / 7))
            except ZeroDivisionError:
                # age could be zero right after channel becomes pending
                sent_received_per_week = 0

            # determine policy:
            try:
                edge_info = self.network.edges[c.chan_id]
                # interested in node2
                policies = edge_info['policies']
                if edge_info['node1_pub'] == self.pub_key:
                    policy_peer = policies[edge_info['node2_pub'] > edge_info['node1_pub']]
                    policy_local = policies[edge_info['node1_pub'] > edge_info['node2_pub']]
                else:  # interested in node1
                    policy_peer = policies[edge_info['node1_pub'] > edge_info['node2_pub']]
                    policy_local = policies[edge_info['node2_pub'] > edge_info['node1_pub']]
            except KeyError:
                # if channel is unknown in describegraph
                # we need to set the fees to some error value
                policy_peer = {
                    'fee_base_msat': float(-999),
                    'fee_rate_milli_msat': float(999)
                }
                policy_local = {
                    'fee_base_msat': float(-999),
                    'fee_rate_milli_msat': float(999)
                }

            # calculate last update (days ago)
            def convert_to_days_ago(timestamp):
                return (time.time() - timestamp) / (60 * 60 * 24)
            try:
                last_update = convert_to_days_ago(
                    self.network.edges[c.chan_id]['last_update'])
                last_update_local = convert_to_days_ago(
                    policy_local['last_update'])
                last_update_peer = convert_to_days_ago(
                    policy_peer['last_update'])
            except (TypeError, KeyError):
                last_update = float('nan')
                last_update_peer = float('nan')
                last_update_local = float('nan')

            # define unbalancedness |ub| large means very unbalanced
            channel_unbalancedness, our_commit_fee = \
                local_balance_to_unbalancedness(
                    c.local_balance, c.capacity, c.commit_fee, c.initiator)
            try:
                uptime_lifetime_ratio = c.uptime / c.lifetime
            except ZeroDivisionError:
                uptime_lifetime_ratio = 0

            channels[c.chan_id] = {
                'active': c.active,
                'age': age_days,
                'alias': self.network.node_alias(c.remote_pubkey),
                'amt_to_balanced': int(
                    channel_unbalancedness * c.capacity / 2 - our_commit_fee),
                'capacity': c.capacity,
                'chan_id': c.chan_id,
                'channel_point': c.channel_point,
                'commit_fee': c.commit_fee,
                'fee_per_kw': c.fee_per_kw,
                'peer_base_fee': policy_peer['fee_base_msat'],
                'peer_fee_rate': policy_peer['fee_rate_milli_msat'],
                'local_base_fee': policy_local['fee_base_msat'],
                'local_fee_rate': policy_local['fee_rate_milli_msat'],
                'local_chan_reserve_sat': c.local_chan_reserve_sat,
                'remote_chan_reserve_sat': c.remote_chan_reserve_sat,
                'initiator': c.initiator,
                'last_update': last_update,
                'last_update_local': last_update_local,
                'last_update_peer': last_update_peer,
                'local_balance': c.local_balance,
                'num_updates': c.num_updates,
                'private': c.private,
                'remote_balance': c.remote_balance,
                'remote_pubkey': c.remote_pubkey,
                'sent_received_per_week': sent_received_per_week,
                'total_satoshis_sent': c.total_satoshis_sent,
                'total_satoshis_received': c.total_satoshis_received,
                'unbalancedness': channel_unbalancedness,
                'uptime': c.uptime,
                'lifetime': c.lifetime,
                'uptime_lifetime_ratio': uptime_lifetime_ratio,
            }
        sorted_dict = OrderedDict(
            sorted(channels.items(), key=lambda x: x[1]['alias']))
        return sorted_dict

    def channel_id_to_node_id(self, open_only=False) -> Dict[int, str]:
        channel_id_to_node_id = {}
        closed_channels = self.get_closed_channels()
        open_channels = self.get_open_channels()
        for cid, c in open_channels.items():
            channel_id_to_node_id[cid] = c['remote_pubkey']
        if not open_only:
            for cid, c in closed_channels.items():
                channel_id_to_node_id[cid] = c['remote_pubkey']
        return channel_id_to_node_id

    def node_id_to_channel_ids(self, open_only=False) -> Dict[str, List[int]]:
        node_channels_mapping = defaultdict(list)
        for cid, nid in self.channel_id_to_node_id(open_only=open_only).items():
            node_channels_mapping[nid].append(cid)
        return node_channels_mapping

    def get_inactive_channels(self):
        """
        Returns all inactive channels.
        :return: dict of channels
        """
        channels = self.get_open_channels(public_only=False, active_only=False)
        return {k: c for k, c in channels.items() if not c['active']}

    def get_all_channels(self, excluded_channels: List[int] = None):
        """
        Returns all active and inactive channels.

        :return: dict of channels
        """
        channels = self.get_open_channels(public_only=False, active_only=False)
        return channels

    def get_unbalanced_channels(self, unbalancedness_greater_than=0.0, excluded_channels: List[int] = None, public_only=True, active_only=True):
        """
        Gets all channels which have an absolute unbalancedness
        (-1...1, -1 for outbound unbalanced, 1 for inbound unbalanced)
        larger than unbalancedness_greater_than.

        :param unbalancedness_greater_than: unbalancedness interval, default returns all channels
        :return: all channels which are more unbalanced than the specified interval
        """
        self.public_active_channels = \
            self.get_open_channels(public_only=public_only, active_only=active_only)
        channels = {
            k: c for k, c in self.public_active_channels.items()
            if abs(c['unbalancedness']) >= unbalancedness_greater_than
        }
        channels = {k: v for k, v in channels.items() if k not in (excluded_channels if excluded_channels else [])}
        return channels

    def get_channel_fee_policies(self):
        """
        Gets the node's channel fee policies for every open channel.
        :return: dict
        """
        feereport = self._rpc.FeeReport(lnd.FeeReportRequest())
        channels = {}
        for fee in feereport.channel_fees:
            channels[fee.channel_point] = {
                'base_fee_msat': fee.base_fee_msat,
                'fee_per_mil': fee.fee_per_mil,
                'fee_rate': fee.fee_rate,
            }
        return channels

    def set_channel_fee_policies(self, channels: dict):
        """
        Sets the node's channel fee policy for every channel.
        :param channels: channel point -> fee policy
        """

        for channel_point, channel_fee_policy in channels.items():
            funding_txid, output_index = channel_point.split(':')
            output_index = int(output_index)

            channel_point = lnd.ChannelPoint(
                funding_txid_str=funding_txid, output_index=output_index)

            update_request = lnd.PolicyUpdateRequest(
                chan_point=channel_point,
                base_fee_msat=channel_fee_policy['base_fee_msat'],
                fee_rate=channel_fee_policy['fee_rate'],
                time_lock_delta=channel_fee_policy['cltv'],
            )
            self._rpc.UpdateChannelPolicy(request=update_request)

    @staticmethod
    def timestamp_from_now(offset_days=0):
        """
        Determines the Unix timestamp from offset_days ago.

        :param offset_days: int
        :return: int, Unix timestamp
        """
        now = datetime.datetime.now()
        then = now - datetime.timedelta(days=offset_days)
        then = time.mktime(then.timetuple())
        return int(then)

    def get_forwarding_events(self, offset_days=300):
        """
        Fetches all forwarding events between now and offset_days ago.

        :param offset_days: int
        :return: lnd fowarding events
        """
        now = self.timestamp_from_now()
        then = self.timestamp_from_now(offset_days)

        forwardings = self._rpc.ForwardingHistory(lnd.ForwardingHistoryRequest(
            start_time=then,
            end_time=now,
            num_max_events=NUM_MAX_FORWARDING_EVENTS))

        events = [{
            'timestamp': f.timestamp,
            'chan_id_in': f.chan_id_in,
            'chan_id_out': f.chan_id_out,
            'amt_in': f.amt_in,
            'amt_in_msat': f.amt_in_msat,
            'amt_out': f.amt_out,
            'amt_out_msat': f.amt_out_msat,
            'fee_msat': f.fee_msat,
            'effective_fee': f.fee_msat / f.amt_in_msat
        } for f in forwardings.forwarding_events]

        return events

    def get_closed_channels(self):
        """
        Fetches all closed channels.

        :return: dict, channel list
        """
        request = lnd.ClosedChannelsRequest()
        closed_channels = self._rpc.ClosedChannels(request)
        closed_channels_dict = {}
        for c in closed_channels.channels:
            closed_channels_dict[c.chan_id] = {
                'channel_point': c.channel_point,
                'chain_hash': c.chain_hash,
                'closing_tx_hash': c.closing_tx_hash,
                'remote_pubkey': c.remote_pubkey,
                'capacity': c.capacity,
                'close_height': c.close_height,
                'settled_balance': c.settled_balance,
                'time_locked_balance': c.time_locked_balance,
                'close_type': c.close_type,
            }
        return closed_channels_dict

    @staticmethod
    def handle_payment_error(payment_error):
        """
        Handles payment errors and determines the failed channel.

        :param payment_error:
        :return: int, channel_id of the failed channel.
        """
        if "TemporaryChannelFailure" in payment_error:
            logger.error("   Encountered temporary channel failure.")
            short_channel_groups = extract_short_channel_id_from_string(payment_error)
            channel_id = convert_short_channel_id_to_channel_id(*short_channel_groups)
            return channel_id

    def queryroute_external(self, source_pubkey, target_pubkey, amt_msat,
                            ignored_nodes=(), ignored_channels={},
                            use_mc=False):
        """
        Queries the lnd node for a route.

        Channels and nodes can be ignored if they failed before.

        :param source_pubkey: source node public key
        :type source_pubkey: str
        :param target_pubkey: target node public key
        :type target_pubkey: str
        :param amt_msat: amount to send in msat
        :type amt_msat: int
        :param ignored_nodes: ignored node pubilc keys for the route
        :type ignored_nodes: list[str]
        :param ignored_channels: ignored channel directions for the route
        :type ignored_channels: dict
        :param use_mc: true if mission control should be used to blacklist
                       channels
        :type use_mc: bool
        :return: route expressed in terms of short channel ids
        :rtype: list[int]
        """
        amt_sat = amt_msat // 1000

        # put safety margin when using mc based routing
        # reason is that routes will not be diverse when sending with a larger
        # amount later on due to fees
        # the fees for the route are somewhat accounted for by the margin
        if use_mc:
            amt_sat = int(amt_sat * 1.02)

        # have a safety max fee in sat
        max_fee = 10000

        # convert ignored nodes to api format
        if ignored_nodes:
            ignored_nodes_api = [bytes.fromhex(n) for n in ignored_nodes]
        else:
            ignored_nodes_api = []

        # convert ignored channels to api format
        if ignored_channels:
            ignored_channels_api = []
            for c, cv in ignored_channels.items():
                direction_reverse = cv['source'] > cv['target']
                ignored_channels_api.append(
                    lnd.EdgeLocator(channel_id=c,
                                    direction_reverse=direction_reverse))
        else:
            ignored_channels_api = []

        logger.debug(f"Ignored for queryroutes: channels: "
                     f"{ignored_channels_api}, nodes: {ignored_nodes_api}")

        request = lnd.QueryRoutesRequest(
            pub_key=target_pubkey,
            amt=amt_sat,
            final_cltv_delta=0,
            fee_limit=lnd.FeeLimit(fixed=max_fee),
            ignored_nodes=ignored_nodes_api,
            ignored_edges=ignored_channels_api,
            source_pub_key=source_pubkey,
            use_mission_control=use_mc,
        )
        try:
            response = self._rpc.QueryRoutes(request)
        except Exception as e:
            if "unable to find a path" in e.details():
                raise NoRoute
            else:
                raise e

        # We give back only one route, as multiple routes will be deprecated
        channel_route = [h.chan_id for h in response.routes[0].hops]

        return channel_route

    def get_node_info(self, pub_key):
        """
        Retrieves information on a node with a specific pub key.

        :param pub_key: node public key
        :type pub_key: str
        :return: node information including all channels
        :rtype: dict
        """
        request = lnd.NodeInfoRequest(pub_key=pub_key, include_channels=True)
        try:
            response = self._rpc.GetNodeInfo(request)
        except _Rendezvous as e:
            if e.details() == "unable to find node":
                logger.info(
                    "LND node has no information about node with pub key %s.",
                    pub_key)
            raise KeyError

        node_info = {
            'alias': response.node.alias,
            'color': response.node.color,
            'channels': response.channels,
            'last_update': response.node.last_update,
            'pub_key': pub_key,
            'num_channels': int(response.num_channels),
            'total_capacity': int(response.total_capacity),  # sat
        }

        addresses = [address.addr for address in response.node.addresses]
        node_info['addresses'] = addresses

        return node_info

    def get_utxos(self) -> List[UTXO]:
        response = self._walletrpc.ListUnspent(
            lndwalletkit.ListUnspentRequest(
                min_confs=1,
                max_confs=100000
            ))
        return [UTXO(
            str(u.outpoint.txid_str),
            int(u.outpoint.output_index),
            int(u.amount_sat),
            AddressType(u.address_type)
        ) for u in response.utxos]

    def print_status(self):
        logger.info("-------- Node status --------")
        if self.total_capacity == 0:
            balancedness_local = 0
            balancedness_remote = 0
        else:
            balancedness_local = self.total_local_balance / self.total_capacity
            balancedness_remote = self.total_remote_balance / self.total_capacity
        logger.info(f"alias: {self.alias}")
        logger.info(f"pub key: {self.pub_key}")
        logger.info(f"blockheight: {self.blockheight}")
        logger.info(f"peers: {self.num_peers}")
        logger.info(f"channels: {self.total_channels}")
        logger.info(f"active channels: {self.total_active_channels}")
        logger.info(f"private channels: {self.total_private_channels}")
        logger.info(f"capacity: {self.total_capacity}")
        logger.info(f"balancedness: l:{balancedness_local:.2%} r:{balancedness_remote:.2%}")
        logger.info(f"total satoshis received (current channels): {self.total_satoshis_received}")
        logger.info(f"total satoshis sent (current channels): {self.total_satoshis_sent}")

    def open_channels(self, pubkeys: List[bytes],
                      amounts_sat: List[int],
                      change_sat: int,
                      utxos: Optional[List[UTXO]],
                      sat_per_vbyte: int,
                      reckless=False,
                      private=False):
        """
        Batch opens channels to other nodes.
        Needs lnd compiled with walletrpc.
        Should be used with lib.openchannels.ChannelOpener.open_channels
        to sanitize method inputs.

        1. Construct PSBT for funding.
        2. Fund and sanity check PSBT.
        3. Verify PSBT by server.
        4. Sign PSBT.
        5. Tell server about signed PSBT.
        6. Publish funding transaction.
        7. Optionally abort funding and free locked utxos.
        """
        # at this stage we assume that we are already connected to the nodes

        utxo_unlocking_needed = True
        psbt_unfunded = None
        utxo_leases = []
        pending_chan_ids = []
        addresses = []

        logger.info(">>> Asking peers for channel opening.")
        time_start = time.time()
        try:
            for n, (pk, amt) in enumerate(zip(pubkeys, amounts_sat)):
                pending_chan_id = os.urandom(32)
                pending_chan_ids.append(pending_chan_id)

                # 1. Construct PSBT for funding.
                if n == 0:  # for the first iteration we don't have a base psbt
                    psbt_shim = lnd.PsbtShim(pending_chan_id=pending_chan_id, no_publish=True)
                    shim = lnd.FundingShim(psbt_shim=psbt_shim)
                else:
                    psbt_shim = lnd.PsbtShim(pending_chan_id=pending_chan_id, base_psbt=psbt_unfunded, no_publish=True)
                    shim = lnd.FundingShim(psbt_shim=psbt_shim)
                request = lnd.OpenChannelRequest(
                    node_pubkey=pk,
                    local_funding_amount=amt,
                    funding_shim=shim,
                    private=private,
                )
                open_channel_stream = self._rpc.OpenChannel(request)

                for chan_response in open_channel_stream:
                    # first response should be that the funding psbt was created:
                    if chan_response.HasField('psbt_fund'):
                        logger.debug(f"   > peer {pk.hex()[:10]}...: got address: {chan_response.psbt_fund.funding_address}")
                        addresses.append(chan_response.psbt_fund.funding_address)
                        psbt_unfunded = chan_response.psbt_fund.psbt
                        break

            # 2. Fund and sanity check PSBT.
            logger.info(f">>> Funding PSBT.")
            inputs = [
                lnd.OutPoint(
                    txid_str=utxo.txid,
                    output_index=utxo.output_index,
                ) for utxo in utxos
            ]
            outputs = [
                (addr, amount) for addr, amount in zip(addresses, amounts_sat)
            ]
            if change_sat:
                change_address = self._rpc.NewAddress(lnd.NewAddressRequest(
                    type=lnd.AddressType.WITNESS_PUBKEY_HASH)).address
                outputs.append((change_address, change_sat))

            fund_psbt = self._walletrpc.FundPsbt(lndwalletkit.FundPsbtRequest(
                raw=lndwalletkit.TxTemplate(
                    inputs=inputs,
                    outputs=outputs,
                ),
                sat_per_vbyte=sat_per_vbyte,
            ))
            utxo_leases = fund_psbt.locked_utxos

            # sanity checks
            if fund_psbt.change_output_index != -1:
                raise Exception("We shouldn't have an internal change output, please report this.")

            num_inputs, num_outputs, psbt_amounts = extract_psbt_inputs_outputs(fund_psbt.funded_psbt)
            logger.debug(f"    given inputs: {utxos}")
            logger.debug(f"    inputs (psbt): {num_inputs}")
            logger.debug(f"    outputs (psbt): {num_outputs}")
            logger.debug(f"    amounts (psbt): {psbt_amounts}")
            assert num_inputs == len(inputs)
            assert num_outputs == len(outputs)
            if change_sat:
                if change_sat in psbt_amounts:
                    psbt_amounts.remove(change_sat)
                else:
                    raise ValueError("Expected change, but couldn't find it.")
            # order of outputs is not clear, comparing sets
            assert set(amounts_sat) == set(psbt_amounts)
            # add back change
            psbt_amounts.append(change_sat)

            fee = sum(utxo.amount_sat for utxo in utxos) - sum(psbt_amounts)
            assert fee < 100000, f'Fee ({fee} sat) unreasonably high? Stopping.'

            logger.info(f">>> WARNING: this is a relatively new feature, so "
                        f"please check the generated PSBT by the following command:")
            logger.info(f'bitcoin-cli decodepsbt "{str(binascii.b2a_base64(fund_psbt.funded_psbt).strip(), "utf-8")}"')
            logger.info(f">>> You have {OPEN_EXPIRY_TIME_MINUTES} minutes from now to decide.\n")
            logger.info("\n>>> Do you want to open the channel(s) (y/n)?")
            if not reckless and not yes_no_question('no'):
                raise InterruptedError("User canceled the process.")
            time_end = time.time()
            if time_end - time_start > OPEN_EXPIRY_TIME_MINUTES * 60:
                raise InterruptedError("Time expired, aborted the channel opening process.")

            # 3. Verify PSBT by server.
            logger.info(f">>> Verifying PSBT.")
            for p in pending_chan_ids:
                response = str(self._rpc.FundingStateStep(
                    lnd.FundingTransitionMsg(
                        psbt_verify=lnd.FundingPsbtVerify(
                            funded_psbt=fund_psbt.funded_psbt,
                            pending_chan_id=p,
                        ),
                    )
                )).strip()
                if response:
                    logger.debug(response)

            # 4. Sign PSBT.
            logger.info(f">>> Signing PSBT.")
            finalize = self._walletrpc.FinalizePsbt(
                lndwalletkit.FinalizePsbtRequest(funded_psbt=fund_psbt.funded_psbt))
            raw_final_tx = finalize.raw_final_tx
            psbt_signed = finalize.signed_psbt
            logger.info(f"    Signed transaction:\n    {raw_final_tx.hex()}")
            logger.info(f"    Signed psbt:\n    {str(binascii.b2a_base64(psbt_signed).strip(), 'utf-8')}")
            logger.info(f"    Final transaction size: {len(raw_final_tx)} bytes")

            # 5. Tell server about signed PSBT.
            for p in pending_chan_ids:
                response = str(self._rpc.FundingStateStep(
                    lnd.FundingTransitionMsg(
                        psbt_finalize=lnd.FundingPsbtFinalize(
                            signed_psbt=psbt_signed,
                            pending_chan_id=p,
                        ),
                    )
                )).strip()
                if response:
                    logger.debug(f"   > Funding step response: {response}")

            # 6. Publish funding transaction.
            logger.info(f">>> Publishing transaction.")
            self._walletrpc.PublishTransaction(lndwalletkit.Transaction(
                tx_hex=raw_final_tx,
                label='lndmanage: batch open'
            ))
            utxo_unlocking_needed = False

        except grpc.RpcError as e:
            logger.info(f"Error: {e}")

        # 7. Optionally abort funding and free locked utxos.
        finally:
            if utxo_unlocking_needed:
                logger.info(">>> Cleaning up.")
                # cancel all funding reservations
                for p in pending_chan_ids:
                    try:
                        self._rpc.FundingStateStep(
                            lnd.FundingTransitionMsg(
                                shim_cancel=lnd.FundingShimCancel(
                                    pending_chan_id=p,
                                ),
                            )
                        )
                    except Exception as e:
                        logger.info(e)
                # unlock coins
                for lease in utxo_leases:
                    try:
                        self._walletrpc.ReleaseOutput(
                            lndwalletkit.ReleaseOutputRequest(
                                id=lease.id,
                                outpoint=lnd.OutPoint(
                                    txid_str=lease.outpoint.txid_str,
                                    output_index=lease.outpoint.output_index,
                                ),
                            )
                        )
                    except Exception as e:
                        logger.info(e)

    def _connect_nodes(self, pubkeys: List[str]) -> List[str]:
        """
        Raises ConnectionRefusedError.
        """
        succeeded_nodes = []
        logger.info(">>> Checking node pubkeys and address information.")
        for pubkey in pubkeys:
            if len(pubkey) != 66:
                raise ValueError(f"pubkey of unknown format {pubkey}")
            info = self.get_node_info(pubkey)
            if not info['addresses']:
                raise ConnectionRefusedError(f"Could not find connection address for {pubkey}.")
        logger.info(">>> Connecting to channel peer candidates.")
        for pubkey in pubkeys:
            info = self.get_node_info(pubkey)
            for address in info['addresses']:
                logger.info(f"    trying to connect to {pubkey}@{address}")
                try:
                    self._rpc.ConnectPeer(
                        lnd.ConnectPeerRequest(
                            addr=lnd.LightningAddress(
                                pubkey=pubkey,
                                host=address,
                            ),
                            perm=False,
                            timeout=20,
                        ))
                    succeeded_nodes.append(pubkey)
                    logger.info("    > connected")
                    break
                except _InactiveRpcError as e:
                    if "already connected" in e.details():
                        succeeded_nodes.append(pubkey)
                        logger.info("    > already connected")
                        break
                    else:
                        logger.info(f"    > error: {e.details()}")
                except Exception as e:
                    logger.exception(e)
                    continue
            else:
                raise ConnectionRefusedError
        return succeeded_nodes

    def pubkey_to_channel_map(self):
        """
        Determines a dict with node pubkeys this node has a channel with, which
        maps to a list of all the channels with the node.

        :return: dictionary of pubkeys with list of channels as value
        :rtype: dict[list]
        """
        channels = self.get_all_channels()

        node_to_channel_map = defaultdict(list)

        for c, cv in channels.items():
            node_to_channel_map[cv['remote_pubkey']].append(c)

        return node_to_channel_map



import unittest

from lndmanage.lib.rating import node_badness


class TestBadness(unittest.TestCase):
    def test_badness(self):
        node_hops = ['A', 'B', 'C', 'D', 'E', 'F', 'G']
        failed_hop = 3
        for ni, _ in enumerate(node_hops):
            print(node_badness(ni, failed_hop))



"""
Creates reports for forwardings, channel opens/closings, onchain activity.
"""
import logging
from datetime import datetime
from lndmanage.lib.ln_utilities import (
    convert_channel_id_to_short_channel_id,
    height_to_timestamp
)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# number of bins for histograms
NUMBER_OF_BINS = 48

# define character scale for histogram representation
# these are Braille characters, which are growing from zero to eight dots per
# character
CHARACTER_SCALE = [
    u'\u2800',
    u'\u2840',
    u'\u28C0',
    u'\u28C4',
    u'\u28E4',
    u'\u28E6',
    u'\u28F6',
    u'\u28F7',
    u'\u28FF',
]


def print_histogram(histogram_bar, unit, max_scale):
    """
    Prints a histogram over one line of text using Braille unicode chars.

    :param histogram_bar: histogram string
    :type histogram_bar: str
    :param unit: represented y unit of histogram
    :type unit: str
    :param max_scale: the maximal y size of a single bin in the histogram
    :type max_scale: int
    """
    logger.info(
        "   activity (" + CHARACTER_SCALE[8] +
        " represents %s %s):", max_scale, unit)
    logger.info("\n   %s\n", histogram_bar)


class Report(object):
    def __init__(self, node, time_start, time_end):
        """
        :param node: LND node interface
        :type node: lndmanage.lib.node.LndNode
        :param time_start: unix timestamp for beginning of analysis
        :type time_start: int
        :param time_end: unix timestamp for end of analysis
        :type time_end: int
        """
        if time_start > time_end:
            raise ValueError("starting time must be earlier than end time")

        self.node = node
        self.time_start = int(time_start)
        self.time_end = int(time_end)
        offset_days = (self.time_end - self.time_start) // 3600 // 24
        self.forwarding_events = self.node.get_forwarding_events(
            offset_days=offset_days)
        self.channel_closings = self.node.get_closed_channels()
        self.channels = self.node.get_all_channels()

    def report(self):
        """
        Prints different subreports on forwardings and channel events.
        """
        date_start = datetime.fromtimestamp(
            self.time_start).strftime('%Y-%m-%d %H:%M')
        date_end = datetime.fromtimestamp(
            self.time_end).strftime('%Y-%m-%d %H:%M')
        logger.info("\nReport from %s to %s\n", date_start, date_end)
        self.report_forwarding_events()
        logger.info("")
        self.report_forwarding_fees()
        logger.info("")
        self.report_forwarding_amounts()
        logger.info("")
        self.report_channel_closings()
        logger.info("")
        self.report_channel_openings()

    def report_forwarding_events(self):
        """
        Reports forwarding events.
        """
        series = self.get_forwarding_event_series()
        time_series = TimeSeries(series, self.time_start, self.time_end)
        histogram_bar, max_scale = time_series.histogram_bar()

        logger.info("Forwardings:")
        if time_series.total_counts:
            print_histogram(histogram_bar, "forwardings", max_scale)

            logger.info("   total forwardings: %s", time_series.total_values)
            logger.info(
                "   forwardings per day: %d",
                time_series.total_values /
                ((self.time_end - self.time_start) / (24 * 3600)))

            logger.info("\n   channels with most outgoing forwardings:")
            sorted_events_by_key = sorted(
                time_series.events_by_key.items(),
                key=lambda x: x[1]['counts'],
                reverse=True
            )
            for c in sorted_events_by_key[:5]:
                logger.info(f"   {c[0]}: {c[1]['values']}")
        else:
            logger.info("   No forwardings during this time frame.")

    def get_forwarding_event_series(self):
        """
        Fetches forwarding events in the format to be used by TimeSeries.

        :return: time series of forwarding events
        :rtype: list[dict]
        """
        series = [
            {
                'timestamp': event['timestamp'],
                'key': event['chan_id_out'],
                'quantity': 1
            }
            for event in self.forwarding_events]
        return series

    def report_forwarding_fees(self):
        """
        Reports on forwarding fees.
        """
        series = self.get_forwarding_fees_series()
        time_series = TimeSeries(series, self.time_start, self.time_end)
        histogram_bar, max_scale = time_series.histogram_bar()

        logger.info("Forwarding fees:")
        if time_series.total_counts:
            print_histogram(histogram_bar, "msat fees", max_scale)
            logger.info(
                "   total forwarding fees: %s msat", time_series.total_values)
            logger.info(
                "   fees per forwarding: %d msat",
                time_series.total_values / sum(time_series.bins_counts))

            logger.info("\n   channels with most fees collected:")
            sorted_events_by_key = sorted(
                time_series.events_by_key.items(),
                key=lambda x: x[1]['values'],
                reverse=True
            )
            for c in sorted_events_by_key[:5]:
                logger.info(f"   {c[0]}: {c[1]['values']} msat")
        else:
            logger.info("   No forwardings during this time frame.")

    def get_forwarding_fees_series(self):
        """
        Fetches forwarding fee series to be used by TimeSeries.

        :return: forwarding fee series
        :rtype: list[dict]
        """
        series = [
            {
                'timestamp': event['timestamp'],
                'key': event['chan_id_out'],
                'quantity': event['fee_msat']
            }
            for event in self.forwarding_events]
        return series

    def report_forwarding_amounts(self):
        """
        Reports forwarding amounts.
        """
        series = self.get_forwarding_amounts_series()
        time_series = TimeSeries(series, self.time_start, self.time_end)
        histogram_bar, max_scale = time_series.histogram_bar()

        logger.info("Forwarding amount:")
        if time_series.total_counts:
            print_histogram(histogram_bar, "sat", max_scale)
            logger.info("   total forwarded: %s sat", time_series.total_values)
            logger.info(
                "   amount per forwarding: %d sat",
                time_series.total_values / sum(time_series.bins_counts))

            logger.info("\n   channels with most forwarding amounts:")
            sorted_events_by_key = sorted(
                time_series.events_by_key.items(),
                key=lambda x: x[1]['values'],
                reverse=True
            )
            for c in sorted_events_by_key[:5]:
                logger.info(f"   {c[0]}: {c[1]['values']} sat")
        else:
            logger.info("   No forwardings during this time frame.")

    def get_forwarding_amounts_series(self):
        """
        Fetches forwarding amount series to be used by TimeSeries.

        :return: forwarding amount series
        :rtype: list[dict]
        """
        series = [
            {
                'timestamp': event['timestamp'],
                'key': event['chan_id_out'],
                'quantity': event['amt_out']
            }
            for event in self.forwarding_events]
        return series

    def report_channel_closings(self):
        """
        Reports channel closings.
        """
        logger.info("Channel closings:")
        series = self.get_channel_closings_series()
        time_series = TimeSeries(series, self.time_start, self.time_end)
        histogram_bar, max_scale = time_series.histogram_bar()
        if time_series.total_counts:
            print_histogram(histogram_bar, "sat", max_scale)
            logger.info("   total closings: %s", sum(time_series.bins_counts))
            logger.info("   freed funds: %s sat", sum(time_series.bins_values))

            logger.info("\n   closed channels:")
            for c in time_series.events_by_key.items():
                logger.info(f"   {c[0]}: {c[1]['values']} sat freed")
        else:
            logger.info("   No channel closings during this time frame.")

    def get_channel_closings_series(self):
        """
        Fetches forwarding amount series to be used by TimeSeries.

        :return: channel closing series
        :rtype: list[dict]
        """

        series = [
            {
                # calculate back, when approximately the channel was closed
                'timestamp': height_to_timestamp(self.node,
                                                 event['close_height']),
                'key': event_key,
                'quantity': event['settled_balance']
            }
            for event_key, event in self.channel_closings.items()]
        return series

    def report_channel_openings(self):
        """
        Reports channel openings.
        """
        logger.info("Channel openings (of current channels):")
        series = self.get_channel_openings_series()
        time_series = TimeSeries(series, self.time_start, self.time_end)
        histogram_bar, max_scale = time_series.histogram_bar()

        if time_series.total_counts:
            print_histogram(
                histogram_bar, "capacity added in sat", max_scale)

            logger.info(
                "   total openings: %s", sum(time_series.bins_counts))
            logger.info(
                "   total capacity added: %s sat",
                sum(time_series.bins_values))

            logger.info("\n   opened channels:")
            sorted_events_by_key = sorted(
                time_series.events_by_key.items(),
                key=lambda x: x[0]
            )
            for c in sorted_events_by_key:
                logger.info(f"   {c[0]}: {c[1]['values']} sat of new capacity")
        else:
            logger.info("   No channel openings during this time frame.")

    def get_channel_openings_series(self):
        """
        Fetches channel opening series to be used by TimeSeries.

        :return: channel opening series
        :rtype: list[dict]
        """
        series = []
        for chan_id, channel_values in self.channels.items():
            blockheight = convert_channel_id_to_short_channel_id(chan_id)[0]
            series.append({
                'timestamp': height_to_timestamp(self.node, blockheight),
                'key': chan_id,
                'quantity': channel_values['capacity']
            })
        return series


class TimeSeries(object):
    """
    Object to calculate time series histograms on data with special format,
    look at the above get_series methods.
    """
    def __init__(self, series, time_start, time_end):
        """
        :param series: data series
        :type series: list[dict]
        :param time_start: unix timestamp
        :type time_start: int
        :param time_end: unix timestamp
        :type time_end: int
        """
        self.series = series
        self.time_start = time_start
        self.time_end = time_end
        self.time_interval_sec = (time_end - time_start) // NUMBER_OF_BINS
        self.binned_series = []
        self.events_by_key = {}

        self.binned_series, self.events_by_key = self.create_binned_series()
        self.bins_values = list(
            map(lambda x: self.sum_data(x['data']), self.binned_series))
        self.bins_counts = list(
            map(lambda x: len(x['data']), self.binned_series))
        self.total_values = sum(self.bins_values)
        self.total_counts = sum(self.bins_counts)

    def create_binned_series(self):
        """
        Creates the histogram and analyzes the data by keys.

        :return: binned series, events by key
        :rtype: (list[dict], dict)
        """
        binned_series = []
        events_by_key = {}

        # initialize bins
        for d in range(NUMBER_OF_BINS):
            binned_series.append({
                'time_start': self.time_start + d * self.time_interval_sec,
                'time_end': self.time_start + (d + 1) * self.time_interval_sec,
                'data': []
            })

        # fill bins with key and quantity
        for s in self.series:
            _bin = (s['timestamp'] - self.time_start) // self.time_interval_sec
            if 0 <= _bin < NUMBER_OF_BINS:
                binned_series[_bin]['data'].append((s['key'], s['quantity']))

                # accumulate keyed values to have additional statistics
                if s['key'] in events_by_key:
                    events_by_key[s['key']]['counts'] += 1
                    events_by_key[s['key']]['values'] += s['quantity']
                else:
                    events_by_key[s['key']] = {
                        'counts': 1, 'values': s['quantity']}
        return binned_series, events_by_key

    def sum_data(self, data):
        """
        Sums up data in data.

        :param data: contains list of events with first index timestamp and
            second index the value of the event
        :type data: list[tuple]
        :return: data sum
        :rtype: int
        """
        s = 0
        for d in data:
            s += d[1]
        return s

    def histogram_bar(self):
        """
        Creates the histogram string from the binned data using Braille chars.

        :return: the histogram string and the maximal value of the bins
        :rtype: (str, int)
        """
        bar = '|'
        max_count = max(self.bins_values)
        # take 8 as the default max scale (Braille chars have eight dots)
        if max_count <= 8:
            max_count = 8
        # normalize to ints of maximal size of eight
        normalized_counts = [
            int(round(8 * c / max_count, 0)) for c in self.bins_values]
        # map ints to Braille characters
        for s in normalized_counts:
            bar += CHARACTER_SCALE[s]
        bar += '|'
        return bar, max_count



"""
Tests for circular self-payments.
"""
import math
import time
from typing import List
import unittest

from lndmanage.lib.listings import ListChannels
from lndmanage.lib.rebalance import Rebalancer
from lndmanage.lib.exceptions import (
    TooExpensive,
    DryRun,
    RebalancingTrialsExhausted,
    NoRoute,
    OurNodeFailure,
)
from lndmanage import settings

from test.testing_common import (
    test_graphs_paths,
    lndmanage_home,
    SLEEP_SEC_AFTER_REBALANCING,
    TestNetwork,
)

import logging.config
settings.set_lndmanage_home_dir(lndmanage_home)
logging.config.dictConfig(settings.logger_config)
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
logger.handlers[0].setLevel(logging.DEBUG)


class CircleTest(TestNetwork):
    """Implements testing of circular self-payments.
    """
    network_definition = None

    def circular_rebalance_and_check(
            self,
            channel_numbers_send: List[int],
            channel_numbers_receive: List[int],
            amount_sat: int,
            expected_fees_msat: int,
            budget_sat=20,
            max_effective_fee_rate=50,
            dry=False
    ):
        """Helper function for testing a circular payment.

        :param channel_numbers_send: channels whose local balance is decreased
        :param channel_numbers_receive: channels whose local balance is increased
        :param amount_sat: amount in satoshi to rebalance
        :param expected_fees_msat: expected fees in millisatoshi for the rebalance
        :param budget_sat: budget for rebalancing
        :param max_effective_fee_rate: the maximal effective fee rate accepted
        :param dry: if it should be a dry run
        """

        self.rebalancer = Rebalancer(
            self.lndnode,
            max_effective_fee_rate=max_effective_fee_rate,
            budget_sat=budget_sat,
            force=True,
        )

        graph_before = self.testnet.assemble_graph()
        send_channels = {}
        self.rebalancer.channels = self.lndnode.get_unbalanced_channels()

        for c in channel_numbers_send:
            channel_id = self.testnet.channel_mapping[c]['channel_id']
            send_channels[channel_id] = self.rebalancer.channels[channel_id]
        receive_channels = {}
        for c in channel_numbers_receive:
            channel_id = self.testnet.channel_mapping[c]['channel_id']
            receive_channels[channel_id] = self.rebalancer.channels[channel_id]

        invoice = self.lndnode.get_invoice(amount_sat, '')
        payment_hash, payment_address = invoice.r_hash, invoice.payment_addr

        fees_msat = self.rebalancer._rebalance(
            send_channels=send_channels,
            receive_channels=receive_channels,
            amt_sat=amount_sat,
            payment_hash=payment_hash,
            payment_address=payment_address,
            budget_sat=budget_sat,
            dry=dry
        )

        time.sleep(SLEEP_SEC_AFTER_REBALANCING)  # needed to let lnd update the balances

        graph_after = self.testnet.assemble_graph()

        self.assertEqual(expected_fees_msat, fees_msat)

        # check that we send the amount we wanted and that it's conserved
        # TODO: this depends on channel reserves, we assume we opened the channels
        sent = 0
        received = 0
        for c in channel_numbers_send:
            sent += (graph_before['A'][c]['local_balance'] - graph_after['A'][c]['local_balance'])
        for c in channel_numbers_receive:
            received += (graph_before['A'][c]['remote_balance'] - graph_after['A'][c]['remote_balance'])
        assert sent - math.ceil(expected_fees_msat / 1000) == received

        listchannels = ListChannels(self.lndnode)
        listchannels.print_all_channels('rev_alias')

    def graph_test(self):
        """
        graph_test should be implemented by each subclass test and check,
        wheteher the test graph has the correct shape.
        """
        raise NotImplementedError


@unittest.skip
class TestCircleLiquid(CircleTest):
    network_definition = test_graphs_paths['star_ring_3_liquid']

    def graph_test(self):
        # assert some basic properties of the graph
        self.assertEqual(6, len(self.master_node_graph_view))

    def test_circle_success_1_2(self):
        """
        Test successful rebalance from channel 1 to channel 2.
        """
        channel_numbers_from = [1]
        channel_numbers_to = [2]
        amount_sat = 10000
        expected_fees_msat = 43

        self.circular_rebalance_and_check(
            channel_numbers_from,
            channel_numbers_to,
            amount_sat,
            expected_fees_msat
        )

    def test_circle_success_1_6(self):
        """
        Test successful rebalance from channel 1 to channel 6.
        """
        channel_numbers_from = [1]
        channel_numbers_to = [6]
        amount_sat = 10000
        expected_fees_msat = 33

        self.circular_rebalance_and_check(
            channel_numbers_from,
            channel_numbers_to,
            amount_sat,
            expected_fees_msat
        )

    def test_circle_6_1_fail_rebalance_failure_no_funds(self):
        """
        Test expected failure for channel 6 to channel 1, where channel 6
        doesn't have funds.
        """
        channel_numbers_from = [6]
        channel_numbers_to = [1]
        amount_sat = 10000
        expected_fees_msat = 33

        self.assertRaises(
            OurNodeFailure,
            self.circular_rebalance_and_check,
            channel_numbers_from,
            channel_numbers_to,
            amount_sat,
            expected_fees_msat,
        )

    def test_circle_1_6_fail_budget_too_expensive(self):
        """
        Test expected failure where rebalance uses more than the fee budget.
        """
        channel_numbers_from = [1]
        channel_numbers_to = [6]
        amount_sat = 10000
        expected_fees_msat = 33
        budget_sat = 0

        self.assertRaises(
            TooExpensive,
            self.circular_rebalance_and_check,
            channel_numbers_from,
            channel_numbers_to,
            amount_sat,
            expected_fees_msat,
            budget_sat,
        )

    def test_circle_1_6_fail_max_fee_rate_too_expensive(self):
        """
        Test expected failure where rebalance is more expensive than
        the desired maximal fee rate.
        """
        channel_numbers_from = [1]
        channel_numbers_to = [6]
        amount_sat = 10000
        expected_fees_msat = 33
        budget = 20
        max_effective_fee_rate = 0

        self.assertRaises(
            TooExpensive,
            self.circular_rebalance_and_check,
            channel_numbers_from,
            channel_numbers_to,
            amount_sat,
            expected_fees_msat,
            budget,
            max_effective_fee_rate,

        )

    def test_circle_1_6_success_channel_reserve(self):
        """
        Test for a maximal amount circular payment.
        """
        channel_numbers_from = [1]
        channel_numbers_to = [6]
        local_balance = 1000000
        # take into account 1% channel reserve
        amount_sat = int(local_balance - 0.01 * local_balance)
        # need to subtract commitment fee (local + anchor output)
        amount_sat -= 3140
        # need to subtract anchor values
        amount_sat -= 330 * 2
        # need to also subtract fees, then error message changes
        amount_sat -= 3
        # extra to make it work for in-between nodes
        amount_sat -= 200
        # TODO: figure out exactly the localbalance - fees for initiator

        expected_fees_msat = 2_959

        self.circular_rebalance_and_check(
            channel_numbers_from,
            channel_numbers_to,
            amount_sat,
            expected_fees_msat,
        )

    def test_circle_1_6_fail_rebalance_dry(self):
        """
        Test if dry run exception is raised.
        """
        channel_numbers_from = [1]
        channel_numbers_to = [6]
        amount_sat = 10000
        expected_fees_msat = 33

        self.assertRaises(
            DryRun,
            self.circular_rebalance_and_check,
            channel_numbers_from,
            channel_numbers_to,
            amount_sat,
            expected_fees_msat,
            dry=True
        )

    @unittest.skip
    def test_multi_send(self):
        pass

    @unittest.skip
    def test_multi_receive(self):
        pass


class TestCircleIlliquid(CircleTest):

    network_definition = test_graphs_paths['star_ring_4_illiquid']

    def graph_test(self):
        self.assertEqual(10, len(self.master_node_graph_view))

    def test_circle_fail_2_3_no_route(self):
        """Test if NoRoute is raised. We can't go beyond C."""
        channel_numbers_from = [2]  # A -> C
        channel_numbers_to = [3]  # D -> A
        amount_sat = 500_000
        expected_fees_msat = None

        self.assertRaises(
            NoRoute,
            self.circular_rebalance_and_check,
            channel_numbers_from,
            channel_numbers_to,
            amount_sat,
            expected_fees_msat
        )

    def test_circle_1_2_fail_max_trials_exhausted(self):
        """Test if RebalancingTrialsExhausted is raised.

        There will be a single rebalancing attempt, which fails, after which we don't retry.
        """
        channel_numbers_from = [1]
        channel_numbers_to = [2]
        amount_sat = 190950
        expected_fees_msat = None
        settings.REBALANCING_TRIALS = 1

        self.assertRaises(
            RebalancingTrialsExhausted,
            self.circular_rebalance_and_check,
            channel_numbers_from,
            channel_numbers_to,
            amount_sat,
            expected_fees_msat
        )

    def test_circle_1_2_fail_no_route_multi_trials(self):
        """Test if NoRoute is raised."""
        channel_numbers_from = [1]
        channel_numbers_to = [2]
        amount_sat = 450000
        expected_fees_msat = None

        self.assertRaises(
            RebalancingTrialsExhausted,
            self.circular_rebalance_and_check,
            channel_numbers_from,
            channel_numbers_to,
            amount_sat,
            expected_fees_msat
        )



"""PSBT (BIP 174) utilities.
https://github.com/bitcoin/bips/blob/master/bip-0174.mediawiki
Sources:
https://github.com/Jason-Les/python-psbt
Programming Bitcoin book, Jimmy Song
"""
from io import BytesIO
from typing import List, Tuple, Optional

PSBT_MAGIC_SEPARATOR = b'psbt' + b'\xFF'
PSBT_GLOBAL_UNSIGNED_TX = b'\x00'


def little_endian_to_int(b: bytes):
    '''little_endian_to_int takes bytes sequence as a little-endian number.
    Returns an integer'''
    # use the from_bytes method of int
    return int.from_bytes(b, 'little')


def read_varint(s: BytesIO):
    '''read_varint reads a variable integer from a stream'''
    i = s.read(1)[0]
    if i == 0xfd:
        # 0xfd means the next two bytes are the number
        return little_endian_to_int(s.read(2))
    elif i == 0xfe:
        # 0xfe means the next four bytes are the number
        return little_endian_to_int(s.read(4))
    elif i == 0xff:
        # 0xff means the next eight bytes are the number
        return little_endian_to_int(s.read(8))
    else:
        # anything else is just the integer
        return i


def parse_key_value(stream: BytesIO) -> Tuple[Optional[bytes], Optional[bytes]]:
    key_length = read_varint(stream)
    # a key length of 0 represents a separator
    if key_length == 0:
        return None, None
    key = stream.read(key_length)
    val_length = read_varint(stream)
    val = stream.read(val_length)
    return key, val


def extract_psbt_inputs_outputs(psbt: bytes) -> Tuple[int, int, List[int]]:
    """Parses only the transaction in the global map of a PSBT representing
    an unsigned transaction and returns the number of inputs, outputs, and the
    individual amounts."""
    stream = BytesIO(psbt)
    # parse header
    header = stream.read(5)
    if header != PSBT_MAGIC_SEPARATOR:
        raise ValueError("wrong psbt header")

    # parse global
    key, value = parse_key_value(stream)
    if key != PSBT_GLOBAL_UNSIGNED_TX:
        raise NotImplementedError("Can't parse PSBTs that contain other data than unsigned transactions.")

    # parse transaction
    transaction_stream = BytesIO(value)
    version = little_endian_to_int(transaction_stream.read(4))

    # parse inputs
    num_inputs = read_varint(transaction_stream)
    for _ in range(num_inputs):
        prev_tx = transaction_stream.read(32)[::-1]
        prev_index = little_endian_to_int(transaction_stream.read(4))
        script_sig_length = read_varint(transaction_stream)
        script_sig = transaction_stream.read(script_sig_length)
        sequence = little_endian_to_int(transaction_stream.read(4))

    # parse outputs
    num_outputs = read_varint(transaction_stream)
    output_amounts = []
    for _ in range(num_outputs):
        amount = little_endian_to_int(transaction_stream.read(8))
        script_pubkey_length = read_varint(transaction_stream)
        script_pubkey = transaction_stream.read(script_pubkey_length)
        output_amounts.append(amount)
    return num_inputs, num_outputs, output_amounts



import os
import shutil
from unittest import TestCase

from lnregtest.lib.network import Network

from lib.node import LndNode
from lndmanage import settings

settings.CACHING_RETENTION_MINUTES = 0

# constants for testing
SLEEP_SEC_AFTER_REBALANCING = 2

# testing base folder
test_dir = os.path.dirname(os.path.realpath(__file__))
bin_dir = os.path.join(test_dir, 'bin')
graph_definitions_dir = os.path.join(test_dir, 'graph_definitions')
test_data_dir = os.path.join(test_dir, 'test_data')

# set lndmanage_home path to be in the test_data folder and make sure the
# folder exists
lndmanage_home = os.path.join(test_data_dir, 'lndmanage')
os.makedirs(lndmanage_home, exist_ok=True)

test_graphs_paths = {
    'star_ring_3_liquid': os.path.join(
        graph_definitions_dir, 'star_ring_3_liquid.py'),
    'star_ring_4_unbalanced': os.path.join(
        graph_definitions_dir, 'star_ring_4_unbalanced.py'),
    'star_ring_4_illiquid': os.path.join(
        graph_definitions_dir, 'star_ring_4_illiquid.py'),
    'empty_graph': os.path.join(
        graph_definitions_dir, 'empty_graph.py'),
}


class TestNetwork(TestCase):
    """
    Class for spinning up simulated Lightning Networks to do integration
    testing.

    The implementation inheriting from this class needs to implement the
    graph_test method, which tests properties specific to the chosen test
    network graph. The attribute network_definition is a string that points
    to the file location of a network graph definition in terms of a dict.
    """
    network_definition = None

    def setUp(self):
        if self.network_definition is None:
            self.skipTest("This class doesn't represent a real test case.")
            raise NotImplementedError("A network definition path needs to be "
                                      "given.")

        # we delete previous test data
        try:
            shutil.rmtree(lndmanage_home)
        except FileNotFoundError:
            pass
        os.mkdir(lndmanage_home)

        self.testnet = Network(
            binary_folder=bin_dir,
            network_definition_location=self.network_definition,
            nodedata_folder=test_data_dir,
            node_limit='H',
            from_scratch=True
        )
        self.testnet.run_nocleanup()
        # to run the lightning network in the background and do some testing
        # here, run:
        # $ lnregtest --nodedata_folder /path/to/lndmanage/test/test_data/
        # self.testnet.run_from_background()

        # logger.info("Generated network information:")
        # logger.info(format_dict(self.testnet.node_mapping))
        # logger.info(format_dict(self.testnet.channel_mapping))
        # logger.info(format_dict(self.testnet.assemble_graph()))

        master_node_data_dir = self.testnet.master_node.data_dir
        master_node_port = self.testnet.master_node._grpc_port

        self.master_node_graph_view = self.testnet.master_node_graph_view()

        self.lndnode = LndNode(
            lnd_home=master_node_data_dir,
            lnd_host='localhost:' + str(master_node_port),
            regtest=True
        )
        self.graph_test()

    def tearDown(self):
        self.testnet.cleanup()
        del self.testnet

    def graph_test(self):
        """
        graph_test should be implemented by each subclass test and check,
        whether the test graph has the correct shape.
        """
        raise NotImplementedError


"""
Implements a complete graph, where the master node A can be thought of
being surrounded by four nodes, which share a liquid network of channels.
The master node has unbalanced total inbound to outbound ratio.
"""
nodes = {
    'A': {
        'grpc_port': 11009,
        'rest_port': 8080,
        'port': 9735,
        'base_fee_msat': 1,
        'fee_rate': 0.000001,
        'channels': {
            1: {
                'to': 'B',
                'capacity': 1000000,
                'ratio_local': 10,
                'ratio_remote': 0,
            },
            2: {
                'to': 'C',
                'capacity': 1000000,
                'ratio_local': 3,
                'ratio_remote': 7,
            },
            3: {
                'to': 'D',
                'capacity': 1000000,
                'ratio_local': 3,
                'ratio_remote': 7,
            },
            4: {
                'to': 'E',
                'capacity': 1000000,
                'ratio_local': 4,
                'ratio_remote': 6,
            },
        }
    },
    'B': {
        'grpc_port': 11010,
        'rest_port': 8081,
        'port': 9736,
        'base_fee_msat': 2,
        'fee_rate': 0.000001,
        'channels': {
            5: {
                'to': 'C',
                'capacity': 5000000,
                'ratio_local': 5,
                'ratio_remote': 5,
            },
            6: {
                'to': 'D',
                'capacity': 5000000,
                'ratio_local': 5,
                'ratio_remote': 5,
            },
            7: {
                'to': 'E',
                'capacity': 5000000,
                'ratio_local': 5,
                'ratio_remote': 5,
            },
        }
    },
    'C': {
        'grpc_port': 11011,
        'rest_port': 8082,
        'port': 9737,
        'base_fee_msat': 2,
        'fee_rate': 0.000001,
        'channels': {
            8: {
                'to': 'D',
                'capacity': 5000000,
                'ratio_local': 5,
                'ratio_remote': 5,
            },
            9: {
                'to': 'E',
                'capacity': 5000000,
                'ratio_local': 5,
                'ratio_remote': 5,
            },
        }
    },
    'D': {
        'grpc_port': 11012,
        'rest_port': 8083,
        'port': 9738,
        'base_fee_msat': 2,
        'fee_rate': 0.000001,
        'channels': {
            10: {
                'to': 'E',
                'capacity': 5000000,
                'ratio_local': 5,
                'ratio_remote': 5,
            },
        }
    },
    'E': {
        'grpc_port': 11013,
        'rest_port': 8084,
        'port': 9739,
        'base_fee_msat': 2,
        'fee_rate': 0.000001,
        'channels': {
        }
    },
}



"""
Integration tests for lndmanage.
"""
from test.testing_common import test_graphs_paths, TestNetwork


class NewNode(TestNetwork):
    """
    NewNode tests behavior of lndmanage under a blank new node without any
    channels.
    """
    network_definition = test_graphs_paths['empty_graph']

    def graph_test(self):
        self.assertEqual(0, len(self.master_node_graph_view))

    def test_empty(self):
        # LND interface of lndmanage is initialized in setUp method of super
        # class, so nothing is needed here.
        pass
